#! /usr/bin/env python

import os
import sys

if os.environ.get('DUCKY_IMPORT_DEVEL', 'no') == 'yes':
  sys.path.insert(0, os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))

try:
  import cPickle as pickle

except ImportError:
  import pickle

import collections
import optparse

import ducky.patch
import ducky.log
import ducky.cpu
import ducky.cpu.instructions
import ducky.mm.binary
import ducky.util

from ducky.mm import ADDR_FMT

def main():
  parser = optparse.OptionParser()

  ducky.util.add_common_options(parser)

  parser.add_option('-i', dest = 'files_in', action = 'append', default = [], help = 'Input file')
  parser.add_option('-b', dest = 'binary', action = 'store', default = None, help = 'Profiled binary')

  group = optparse.OptionGroup(parser, 'Output options')
  parser.add_option_group(group)
  group.add_option('-n', '--num-entries', dest = 'num_entries', action = 'store', type = int, default = 20, metavar = 'N', help = 'Display N entries in any listing')

  options, logger = ducky.util.parse_options(parser)

  if not options.files_in:
    parser.print_help()
    sys.exit(1)

  if not options.binary:
    parser.print_help()
    sys.exit(1)

  data = collections.defaultdict(int)

  for path in options.files_in:
    logger.info('Reading profile data from %s', path)

    with open(path, 'rb') as f:
      d = pickle.load(f)

      for k, v in d.iteritems():
        data[k] += v

  all_hits = sum(data.itervalues())

  binary = ducky.mm.binary.File(logger, options.binary, 'r')
  binary.load()
  binary.load_symbols()

  symbol_table = ducky.util.SymbolTable(binary)

  def print_points(addresses):
    table = [
      ['Address', 'Symbol', 'Hits', 'Percentage', 'Inst']
    ]

    for addr in addresses:
      hits = data[addr]
      symbol, offset = symbol_table[addr]

      symbol_name = symbol + (('[%s]' % offset) if offset is not None else '')

      symbol = symbol_table.get_symbol(symbol)
      header, content = binary.get_section(symbol.section)

      table.append([ADDR_FMT(addr), symbol_name, hits, '%.02f' % (float(hits) / float(all_hits) * 100.0), ducky.cpu.instructions.DuckyInstructionSet.decode_instruction(content[(symbol.address + offset - header.base) / 4])])

    logger.table(table)

  def print_symbol_hits(functions):
    table = [
      ['Symbol', 'Hits', 'Percentage']
    ]

    for fn in functions:
      hits = symbol_hits[fn]
      table.append([fn, hits, '%.02f' % (float(hits) / float(all_hits) * 100.0)])

    logger.table(table)

  symbol_hits = collections.defaultdict(int)

  for addr, hits in data.iteritems():
    hits = data[addr]
    symbol, offset = symbol_table[addr]

    symbol_hits[symbol] += hits

  logger.info('')

  keys = sorted(data.keys(), key = data.get, reverse = True)
  print_points(keys[0:options.num_entries])

  logger.info('')

  keys = sorted(symbol_hits.keys(), key = symbol_hits.get, reverse = True)
  print_symbol_hits(keys[0:options.num_entries])

if __name__ == '__main__':
  main()



import ctypes
import enum
import re

from six import integer_types, string_types, add_metaclass
from six.moves import range

from ctypes import LittleEndianStructure, c_uint, c_int

from .registers import Registers, REGISTER_NAMES
from ..mm import OFFSET_FMT, i16, u16, UInt32
from ..util import str2int

class GenericInstBinaryFormat_Overall(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('u32', c_uint)
  ]

class GenericInstBinaryFormat_Opcode(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('opcode', c_uint, 6),
    ('__padding__', c_uint, 26)
  ]

PO_REGISTER  = r'(?P<register_n{operand_index}>(?:r\d\d?)|(?:sp)|(?:fp)|(?:ds))'
PO_AREGISTER = r'(?P<address_register>r\d\d?|sp|fp|ds)(?:(?P<pointer>\[|\()(?:(?P<immediate_sign>-|\+)?(?P<offset_immediate>0x[0-9a-fA-F]+|\d+)|(?P<offset_register>r\d\d?|sp|fp|ds))(?:\]|\)))?'
PO_IMMEDIATE = r'(?:(?P<immediate>(?:-|\+)?(?:0x[0-9a-fA-F]+|\d+))|(?P<immediate_address>&[a-zA-Z_\.][a-zA-Z0-9_]*))'


def BF_FLG(n):
  return '{}:1'.format(n)

def BF_REG(*args):
  args = args or ['reg']
  return '{}:4'.format(args[0])

def BF_IMM(*args):
  args = args or ['immediate']
  return '{}:17:int'.format(args[0])

def BF_IMM_SHORT(*args):
  args = args or ['immediate']
  return '{}:12:int'.format(args[0])

class InstDescriptor(object):
  mnemonic      = None
  opcode        = None
  operands      = None
  binary_format = None

  pattern       = None
  binary_format_name = None

  relative_address = False

  def create_binary_format_class(self):
    fields = []

    fields_desc = self.binary_format if self.binary_format else []

    if not len(fields_desc) or not fields_desc[0].startswith('opcode'):
      fields_desc.insert(0, 'opcode:6')

    for field in fields_desc:
      field = field.split(':')
      data_type = c_int if len(field) == 3 else c_uint
      fields.append((field[0], data_type, int(field[1])))

    self.binary_format_name = 'InstBinaryFormat_{}'.format(self.mnemonic)
    self.binary_format = type(self.binary_format_name, (ctypes.LittleEndianStructure,), {'_pack_': 0, '_fields_': fields})

    def __repr__(__inst):
      return '<' + self.instruction_set.disassemble_instruction(__inst) + '>'

    self.binary_format.__repr__ = __repr__

  def __init__(self, instruction_set):
    super(InstDescriptor, self).__init__()

    self.instruction_set = instruction_set

    pattern = r'\s*' + self.mnemonic

    if self.operands:
      operand_patterns = []

      self.operands = [ot.strip() for ot in self.operands.split(',')]

      for operand_index, operand_types in zip(list(range(0, len(self.operands))), self.operands):
        operand_pattern = []

        for operand_type in operand_types:
          if operand_type == 'r':
            operand_pattern.append(PO_REGISTER.format(operand_index = operand_index))

          elif operand_type == 'a':
            operand_pattern.append(PO_AREGISTER)

          elif operand_type == 'i':
            operand_pattern.append(PO_IMMEDIATE)

          else:
            raise Exception('Unhandled operand type: {}'.format(operand_type))

        operand_patterns.append('(?:' + '|'.join(operand_pattern) + ')')

      pattern += r' ' + ', '.join(operand_patterns)

    pattern = r'^' + pattern + '(?:\s*[;#].*)?$'
    self.pattern = re.compile(pattern, re.MULTILINE)

    self.create_binary_format_class()

    self.instruction_set.instructions.append(self)

  @staticmethod
  def execute(core, inst):
    pass

  @staticmethod
  def assemble_operands(logger, inst, operands):
    pass

  @staticmethod
  def fill_reloc_slot(logger, inst, slot):
    raise NotImplementedError('Instruction descriptor does not support relocation')

  @staticmethod
  def disassemble_operands(inst):
    return []

  @classmethod
  def disassemble_mnemonic(cls, inst):
    return cls.mnemonic

  def emit_instruction(self, logger, line):
    DEBUG = logger.debug

    DEBUG('emit_instruction: input line: %s', line)

    master = self.instruction_set.binary_format_master()
    master.overall.u16 = 0

    DEBUG('emit_instruction: binary format is %s', self.binary_format_name)
    DEBUG('emit_instruction: desc is %s', self)

    real = getattr(master, self.binary_format_name)
    real.opcode = self.opcode

    raw_match = self.pattern.match(line)
    matches = raw_match.groupdict()
    DEBUG('emit_instruction: matches=%s', matches)

    operands = {}

    def str2reg(r):
      if r == 'sp':
        return Registers.SP
      if r == 'ds':
        return Registers.DS
      if r == 'fp':
        return Registers.FP
      return Registers(int(r[1:]))

    if self.operands and len(self.operands):
      for operand_index in range(0, len(self.operands)):
        reg_group_name = 'register_n{}'.format(operand_index)

        if reg_group_name in matches and matches[reg_group_name]:
          operands[reg_group_name] = str2reg(matches[reg_group_name])

        elif 'address_register' in matches and matches['address_register']:
          operands['areg'] = str2reg(matches['address_register'])

          if 'pointer' in matches and matches['pointer'] is not None:
            if matches['pointer'] == '[':
              operands['pointer'] = 'offset'

            elif matches['pointer'] == '(':
              operands['pointer'] = 'segment'

            else:
              raise Exception('Unhandled pointer type: {}'.format(matches))

          if 'offset_register' in matches and matches['offset_register'] is not None:
            operands['offset_register'] = str2reg(matches['offset_register'])

          elif 'offset_immediate' in matches and matches['offset_immediate'] is not None:
            k = -1 if 'immediate_sign' in matches and matches['immediate_sign'] and matches['immediate_sign'].strip() == '-' else 1
            operands['offset_immediate'] = k * str2int(matches['offset_immediate'])

        elif 'immediate' in matches and matches['immediate']:
          operands['immediate'] = str2int(matches['immediate'])

        elif 'immediate_address' in matches and matches['immediate_address'] is not None:
          operands['immediate'] = matches['immediate_address']

        else:
          raise Exception('Unhandled operand: {}'.format(matches))

    else:
      pass

    self.assemble_operands(logger, real, operands)

    for flag in [f for f in dir(self) if f.startswith('flag_')]:
      setattr(real, flag.split('_')[1], getattr(self, flag))

    return real


class InstDescriptor_Generic(InstDescriptor):
  pass

class InstDescriptor_Generic_Unary_R(InstDescriptor):
  operands = 'r'
  binary_format = [BF_REG()]

  @staticmethod
  def assemble_operands(logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    inst.reg = operands['register_n0']

  @staticmethod
  def disassemble_operands(inst):
    return [REGISTER_NAMES[inst.reg]]


class InstDescriptor_Generic_Unary_I(InstDescriptor):
  operands      = 'i'
  binary_format = [BF_IMM()]

  @staticmethod
  def assemble_operands(logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    v = operands['immediate']

    if isinstance(v, integer_types):
      inst.immediate = v

    elif isinstance(v, string_types):
      inst.refers_to = v

  @staticmethod
  def disassemble_operands(inst):
    return [inst.refers_to if hasattr(inst, 'refers_to') and inst.refers_to else OFFSET_FMT(inst.immediate)]

class InstDescriptor_Generic_Unary_RI(InstDescriptor):
  operands      = 'ri'
  binary_format = [BF_FLG('is_reg'), BF_REG('ireg'), BF_IMM()]

  @staticmethod
  def assemble_operands(logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    if 'register_n0' in operands:
      inst.ireg = operands['register_n0']
      inst.is_reg = 1

    else:
      v = operands['immediate']

      inst.is_reg = 0

      if isinstance(v, integer_types):
        inst.immediate = v

      elif isinstance(v, string_types):
        inst.refers_to = v

  @staticmethod
  def fill_reloc_slot(logger, inst, slot):
    logger.debug('fill_reloc_slot: inst=%s, slot=%s', inst, slot)

    slot.patch_offset = 11
    slot.patch_size = 17

  @staticmethod
  def disassemble_operands(inst):
    if inst.is_reg == 1:
      return [REGISTER_NAMES[inst.ireg]]

    return [inst.refers_to if hasattr(inst, 'refers_to') and inst.refers_to else OFFSET_FMT(inst.immediate)]

class InstDescriptor_Generic_Binary_R_I(InstDescriptor):
  operands = 'r,i'
  binary_format = [BF_REG(), BF_IMM()]

  @staticmethod
  def assemble_operands(logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    inst.reg = operands['register_n0']

    v = operands['immediate']
    if isinstance(v, integer_types):
      inst.immediate = v

    else:
      inst.refers_to = v

  @staticmethod
  def fill_reloc_slot(logger, inst, slot):
    logger.debug('fill_reloc_slot: inst=%s, slot=%s', inst, slot)

    slot.patch_offset = 10
    slot.patch_size = 17

  @staticmethod
  def disassemble_operands(inst):
    return [REGISTER_NAMES[inst.reg], inst.refers_to] if hasattr(inst, 'refers_to') and inst.refers_to else [REGISTER_NAMES[inst.reg], OFFSET_FMT(inst.immediate)]

class InstDescriptor_Generic_Binary_R_RI(InstDescriptor):
  operands = 'r,ri'
  binary_format = [BF_FLG('is_reg'), BF_REG(), BF_REG('ireg'), BF_IMM()]

  @staticmethod
  def assemble_operands(logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    inst.reg = operands['register_n0']

    if 'register_n1' in operands:
      inst.ireg = operands['register_n1']
      inst.is_reg = 1

    else:
      inst.is_reg = 0

      v = operands['immediate']

      if isinstance(v, integer_types):
        inst.immediate = v

      elif isinstance(v, string_types):
        inst.refers_to = v

  @staticmethod
  def fill_reloc_slot(logger, inst, slot):
    logger.debug('fill_reloc_slot: inst=%s, slot=%s', inst, slot)

    slot.patch_offset = 15
    slot.patch_size = 17

  @staticmethod
  def disassemble_operands(inst):
    if inst.is_reg == 1:
      return [REGISTER_NAMES[inst.reg], REGISTER_NAMES[inst.ireg]]

    return [REGISTER_NAMES[inst.reg], inst.refers_to if hasattr(inst, 'refers_to') and inst.refers_to else OFFSET_FMT(inst.immediate)]

class InstDescriptor_Generic_Binary_RI_R(InstDescriptor):
  operands = 'ri,r'
  binary_format = [BF_FLG('is_reg'), BF_REG(), BF_REG('ireg'), BF_IMM()]

  @staticmethod
  def assemble_operands(logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    inst.reg = operands['register_n1']

    if 'register_n0' in operands:
      inst.ireg = operands['register_n0']
      inst.is_reg = 1

    else:
      inst.is_reg = 0

      v = operands['immediate']

      if isinstance(v, integer_types):
        inst.immediate = v

      elif isinstance(v, string_types):
        inst.refers_to = v

  @staticmethod
  def disassemble_operands(inst):
    if inst.is_reg == 1:
      return [REGISTER_NAMES[inst.ireg], REGISTER_NAMES[inst.reg]]

    return [inst.refers_to if hasattr(inst, 'refers_to') and inst.refers_to else OFFSET_FMT(inst.immediate), REGISTER_NAMES[inst.reg]]

class InstDescriptor_Generic_Binary_R_A(InstDescriptor):
  operands = 'r,a'
  binary_format = [BF_FLG('is_reg'), BF_FLG('is_segment'), BF_REG(), BF_REG('areg'), BF_REG('oreg'), BF_IMM_SHORT()]

  @staticmethod
  def assemble_operands(logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    inst.reg = operands['register_n0']
    inst.areg = operands['areg']

    inst.is_reg = 0
    inst.is_segment = 0

    if 'offset_register' in operands:
      inst.oreg = operands['offset_register']
      inst.is_reg = 1

    elif 'offset_immediate' in operands:
      inst.immediate = operands['offset_immediate']

    if 'pointer' in operands and operands['pointer'] == 'segment':
      inst.is_segment = 1

  @staticmethod
  def disassemble_operands(inst):
    operands = [REGISTER_NAMES[inst.reg]]

    if inst.is_reg == 1:
      p = '()' if inst.is_segment == 1 else '[]'
      operands.append('{}{}{}{}'.format(REGISTER_NAMES[inst.areg], p[0], REGISTER_NAMES[inst.oreg], p[1]))

    else:
      if inst.immediate == 0:
        operands.append(REGISTER_NAMES[inst.areg])

      else:
        s = '-' if inst.immediate < 0 else ''
        p = '()' if inst.is_segment == 1 else '[]'
        operands.append('{}{}{}{}{}'.format(REGISTER_NAMES[inst.areg], p[0], s, abs(inst.immediate), p[1]))

    return operands

class InstDescriptor_Generic_Binary_A_R(InstDescriptor):
  operands = 'a,r'
  binary_format = [BF_FLG('is_reg'), BF_FLG('is_segment'), BF_REG(), BF_REG('areg'), BF_REG('oreg'), BF_IMM_SHORT()]

  @staticmethod
  def assemble_operands(logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    inst.reg = operands['register_n1']
    inst.areg = operands['areg']

    inst.is_reg = 0
    inst.is_segment = 0

    if 'offset_register' in operands:
      inst.oreg = operands['offset_register']
      inst.is_reg = 1

    elif 'offset_immediate' in operands:
      inst.immediate = operands['offset_immediate']

    if 'pointer' in operands and operands['pointer'] == 'segment':
      inst.is_segment = 1

  @staticmethod
  def disassemble_operands(inst):
    operands = []

    if inst.reg == 1:
      p = '()' if inst.is_segment == 1 else '[]'
      operands.append('{}{}{}{}'.format(REGISTER_NAMES[inst.areg], p[0], REGISTER_NAMES[inst.oreg], p[1]))

    else:
      if inst.immediate == 0:
        operands.append(REGISTER_NAMES[inst.areg])

      else:
        s = '-' if inst.immediate < 0 else ''
        p = '()' if inst.is_segment == 1 else '[]'
        operands.append('{}{}{}{}{}'.format(REGISTER_NAMES[inst.areg], p[0], s, abs(inst.immediate), p[1]))

    operands.append(REGISTER_NAMES[inst.reg])

    return operands

class InstDescriptor_Generic_Binary_R_R(InstDescriptor):
  operands = 'r,r'
  binary_format = [BF_REG('reg1'), BF_REG('reg2')]

  @staticmethod
  def assemble_operands(logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    inst.reg1 = operands['register_n0']
    inst.reg2 = operands['register_n1']

  @staticmethod
  def disassemble_operands(inst):
    return [REGISTER_NAMES[inst.reg1], REGISTER_NAMES[inst.reg2]]

class InstructionSetMetaclass(type):
  def __init__(cls, name, bases, dict):
    cls.instructions = []

@add_metaclass(InstructionSetMetaclass)
class InstructionSet(object):
  instruction_set_id = None
  opcodes = None

  @classmethod
  def init(cls):
    if hasattr(cls, 'opcode_desc_map'):
      return

    cls.opcode_desc_map = {}

    for desc in cls.instructions:
      cls.opcode_desc_map[desc.opcode] = desc

    fields = [
      ('overall', GenericInstBinaryFormat_Overall),
      ('opcode',  GenericInstBinaryFormat_Opcode)
    ]

    for desc in cls.instructions:
      fields.append((desc.binary_format_name, desc.binary_format))

    cls.binary_format_master = type('InstBinaryFormat_Master_{}'.format(cls.__name__), (ctypes.Union,), {'_pack_': 0, '_fields_': fields})

  @classmethod
  def convert_to_master(cls, inst):
    if isinstance(inst, cls.binary_format_master):
      return inst

    master = cls.binary_format_master()
    setattr(master, inst.__class__.__name__, inst)
    return master

  @classmethod
  def decode_instruction(cls, inst):
    if isinstance(inst, integer_types):
      master = cls.binary_format_master()
      master.overall.u32 = inst
      inst = master

    elif isinstance(inst, UInt32):
      master = cls.binary_format_master()
      master.overall.u32 = inst.u32
      inst = master

    if isinstance(inst, cls.binary_format_master):
      if inst.opcode.opcode not in cls.opcode_desc_map:
        from ..cpu import InvalidOpcodeError
        raise InvalidOpcodeError(inst.opcode.opcode)

      return getattr(inst, cls.opcode_desc_map[inst.opcode.opcode].binary_format_name)

    return inst

  @classmethod
  def disassemble_instruction(cls, inst):
    inst = cls.decode_instruction(inst)
    desc = cls.opcode_desc_map[inst.opcode]

    mnemonic = desc.disassemble_mnemonic(inst)
    operands = desc.disassemble_operands(inst)

    return (mnemonic + ' ' + ', '.join(operands)) if operands else mnemonic

#
# Main instruction set
#

class DuckyOpcodes(enum.IntEnum):
  NOP    =  0

  LW     =  1
  LB     =  2
  #      =  3
  LI     =  4
  STW    =  5
  STB    =  6
  MOV    =  8
  SWP    =  9
  CAS    = 10

  INT    = 11
  RETINT = 12

  CALL   = 13
  RET    = 14

  CLI    = 15
  STI    = 16
  RST    = 17
  HLT    = 18
  IDLE   = 19

  PUSH   = 20
  POP    = 21

  INC    = 22
  DEC    = 23
  ADD    = 24
  SUB    = 25
  MUL    = 26

  AND    = 27
  OR     = 28
  XOR    = 29
  NOT    = 30
  SHIFTL = 31
  SHIFTR = 32

  OUT    = 33
  IN     = 34
  OUTB   = 35
  INB    = 36

  CMP    = 37
  J      = 38
  DIV    = 49
  MOD    = 50

  UDIV   = 52

  CMPU   = 51

  # B* instructions
  BRANCH = 39
  BE     = 39
  BNE    = 39
  BZ     = 39
  BNZ    = 39
  BO     = 39
  BNO    = 39
  BS     = 39
  BNS    = 39
  BG     = 39
  BGE    = 39
  BL     = 39
  BLE    = 39

  # SET* instructions
  SET    =  7  # master opcode
  SETE   =  7
  SETNE  =  7
  SETZ   =  7
  SETNZ  =  7
  SETO   =  7
  SETNO  =  7
  SETS   =  7
  SETNS  =  7
  SETL   =  7
  SETLE  =  7
  SETG   =  7
  SETGE  =  7

  SIS    = 63


class Inst_NOP(InstDescriptor_Generic):
  mnemonic = 'nop'
  opcode = DuckyOpcodes.NOP


#
# Interrupts
#
class Inst_INT(InstDescriptor_Generic_Unary_RI):
  mnemonic      = 'int'
  opcode        = DuckyOpcodes.INT

  @staticmethod
  def execute(core, inst):
    core.do_int(core.RI_VAL(inst))

class Inst_RETINT(InstDescriptor_Generic):
  mnemonic = 'retint'
  opcode   = DuckyOpcodes.RETINT

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.exit_interrupt()

#
# Routines
#
class Inst_CALL(InstDescriptor_Generic_Unary_RI):
  mnemonic      = 'call'
  opcode        = DuckyOpcodes.CALL
  relative_address = True

  @staticmethod
  def execute(core, inst):
    core.create_frame()

    core.JUMP(inst)

    if core.check_frames:
      core.frames[-1].IP = core.registers.ip.value

class Inst_RET(InstDescriptor_Generic):
  mnemonic      = 'ret'
  opcode        = DuckyOpcodes.RET

  @staticmethod
  def execute(core, inst):
    core.destroy_frame()

#
# CPU
#
class Inst_CLI(InstDescriptor_Generic):
  mnemonic = 'cli'
  opcode = DuckyOpcodes.CLI

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.registers.flags.hwint = 0

class Inst_STI(InstDescriptor_Generic):
  mnemonic = 'sti'
  opcode = DuckyOpcodes.STI

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.registers.flags.hwint = 1

class Inst_HLT(InstDescriptor_Generic_Unary_RI):
  mnemonic = 'hlt'
  opcode = DuckyOpcodes.HLT

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.exit_code = core.RI_VAL(inst)
    core.halt()

class Inst_RST(InstDescriptor_Generic):
  mnemonic = 'rst'
  opcode = DuckyOpcodes.RST

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.reset()

class Inst_IDLE(InstDescriptor_Generic):
  mnemonic = 'idle'
  opcode = DuckyOpcodes.IDLE

  @staticmethod
  def execute(core, inst):
    core.change_runnable_state(idle = True)

class Inst_SIS(InstDescriptor_Generic_Unary_I):
  mnemonic = 'sis'
  opcode = DuckyOpcodes.SIS

  @staticmethod
  def execute(core, inst):
    core.instruction_set = get_instruction_set(inst.immediate)

#
# Stack
#
class Inst_PUSH(InstDescriptor_Generic_Unary_RI):
  mnemonic = 'push'
  opcode = DuckyOpcodes.PUSH

  @staticmethod
  def execute(core, inst):
    core.raw_push(core.RI_VAL(inst))

class Inst_POP(InstDescriptor_Generic_Unary_R):
  mnemonic = 'pop'
  opcode = DuckyOpcodes.POP

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.pop(inst.reg)
    core.update_arith_flags(core.registers.map[inst.reg])

#
# Arithmetic
#
class Inst_INC(InstDescriptor_Generic_Unary_R):
  mnemonic = 'inc'
  opcode = DuckyOpcodes.INC

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value += 1
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_DEC(InstDescriptor_Generic_Unary_R):
  mnemonic = 'dec'
  opcode = DuckyOpcodes.DEC

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value -= 1
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_ADD(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'add'
  opcode = DuckyOpcodes.ADD

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    v = core.registers.map[inst.reg].value + core.RI_VAL(inst)
    core.registers.map[inst.reg].value += core.RI_VAL(inst)
    core.update_arith_flags(core.registers.map[inst.reg])

    if v > 0xFFFF:
      core.registers.flags.o = 1

class Inst_SUB(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'sub'
  opcode = DuckyOpcodes.SUB

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    v = core.RI_VAL(inst) > core.registers.map[inst.reg].value
    core.registers.map[inst.reg].value -= core.RI_VAL(inst)
    core.update_arith_flags(core.registers.map[inst.reg])

    if v:
      core.registers.flags.s = 1

#
# Conditional and unconditional jumps
#
class InstDescriptor_COND(InstDescriptor):
  binary_format = ['flag:3', BF_FLG('value')]

  FLAGS = ['e', 'z', 'o', 's', 'l', 'g']
  GFLAGS = [0, 1, 2, 3]

  @staticmethod
  def set_condition(inst, flag, value):
    inst.flag = InstDescriptor_COND.FLAGS.index(flag)
    inst.value = 1 if value is True else 0

  @staticmethod
  def evaluate(core, inst):
    # genuine flags
    if inst.flag in InstDescriptor_COND.GFLAGS and inst.value == getattr(core.registers.flags, InstDescriptor_SET.FLAGS[inst.flag]):
      return True

    # "less than" flag
    if inst.flag == 4:
      if inst.value == 1 and core.registers.flags.s == 1 and core.registers.flags.e == 0:
        return True

      if inst.value == 0 and (core.registers.flags.s == 0 or core.registers.flags.e == 1):
        return True

    # "greater than" flag
    if inst.flag == 5:
      if inst.value == 1 and core.registers.flags.s == 0 and core.registers.flags.e == 0:
        return True

      if inst.value == 0 and (core.registers.flags.s == 1 or core.registers.flags.e == 1):
        return True

    return False

class InstDescriptor_BRANCH(InstDescriptor_COND):
  operands = 'ri'
  binary_format = InstDescriptor_COND.binary_format + [BF_FLG('is_reg'), BF_REG('ireg'), BF_IMM()]
  relative_address = True

  @classmethod
  def assemble_operands(cls, logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    if 'register_n0' in operands:
      inst.ireg = operands['register_n0']
      inst.is_reg = 1

    else:
      v = operands['immediate']

      inst.is_reg = 0

      if isinstance(v, integer_types):
        inst.immediate = v

      elif isinstance(v, string_types):
        inst.refers_to = v

    if cls is Inst_BE:
      InstDescriptor_COND.set_condition(inst, 'e', True)

    elif cls is Inst_BNE:
      InstDescriptor_COND.set_condition(inst, 'e', False)

    elif cls is Inst_BZ:
      InstDescriptor_COND.set_condition(inst, 'z', True)

    elif cls is Inst_BNZ:
      InstDescriptor_COND.set_condition(inst, 'z', False)

    elif cls is Inst_BO:
      InstDescriptor_COND.set_condition(inst, 'o', True)

    elif cls is Inst_BNO:
      InstDescriptor_COND.set_condition(inst, 'o', False)

    elif cls is Inst_BS:
      InstDescriptor_COND.set_condition(inst, 's', True)

    elif cls is Inst_BNS:
      InstDescriptor_COND.set_condition(inst, 's', False)

    elif cls is Inst_BL:
      InstDescriptor_COND.set_condition(inst, 'l', True)

    elif cls is Inst_BLE:
      InstDescriptor_COND.set_condition(inst, 'g', False)

    elif cls is Inst_BG:
      InstDescriptor_COND.set_condition(inst, 'g', True)

    elif cls is Inst_BGE:
      InstDescriptor_COND.set_condition(inst, 'l', False)

  @staticmethod
  def fill_reloc_slot(logger, inst, slot):
    logger.debug('fill_reloc_slot: inst=%s, slot=%s', inst, slot)

    slot.patch_offset = 15
    slot.patch_size = 17

  @staticmethod
  def disassemble_operands(inst):
    if inst.is_reg == 1:
      return [REGISTER_NAMES[inst.ireg]]

    return [inst.refers_to if hasattr(inst, 'refers_to') and inst.refers_to else OFFSET_FMT(inst.immediate)]

  @staticmethod
  def disassemble_mnemonic(inst):
    if inst.flag in InstDescriptor_COND.GFLAGS:
      return 'b%s%s' % ('n' if inst.value == 0 else '', InstDescriptor_SET.FLAGS[inst.flag])

    else:
      if inst.flag == InstDescriptor_COND.FLAGS.index('l'):
        return 'bl' if inst.value == 1 else 'bge'

      elif inst.flag == InstDescriptor_COND.FLAGS.index('g'):
        return 'bg' if inst.value == 1 else 'ble'

  @staticmethod
  def execute(core, inst):
    if InstDescriptor_COND.evaluate(core, inst):
      core.JUMP(inst)

class InstDescriptor_SET(InstDescriptor_COND):
  operands = 'r'
  binary_format = InstDescriptor_COND.binary_format + [BF_REG('reg')]

  @classmethod
  def assemble_operands(cls, logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    inst.reg = operands['register_n0']

    if cls is Inst_SETE:
      InstDescriptor_COND.set_condition(inst, 'e', True)

    elif cls is Inst_SETNE:
      InstDescriptor_COND.set_condition(inst, 'e', False)

    elif cls is Inst_SETZ:
      InstDescriptor_COND.set_condition(inst, 'z', True)

    elif cls is Inst_SETNZ:
      InstDescriptor_COND.set_condition(inst, 'z', False)

    elif cls is Inst_SETO:
      InstDescriptor_COND.set_condition(inst, 'o', True)

    elif cls is Inst_SETNO:
      InstDescriptor_COND.set_condition(inst, 'o', False)

    elif cls is Inst_SETS:
      InstDescriptor_COND.set_condition(inst, 's', True)

    elif cls is Inst_SETNS:
      InstDescriptor_COND.set_condition(inst, 's', False)

    elif cls is Inst_SETL:
      InstDescriptor_COND.set_condition(inst, 'l', True)

    elif cls is Inst_SETLE:
      InstDescriptor_COND.set_condition(inst, 'g', False)

    elif cls is Inst_SETG:
      InstDescriptor_COND.set_condition(inst, 'g', True)

    elif cls is Inst_SETGE:
      InstDescriptor_COND.set_condition(inst, 'l', False)

  @staticmethod
  def disassemble_operands(inst):
    return [REGISTER_NAMES[inst.reg]]

  @staticmethod
  def disassemble_mnemonic(inst):
    if inst.flag in InstDescriptor_COND.GFLAGS:
      return 'set%s%s' % ('n' if inst.value == 0 else '', InstDescriptor_SET.FLAGS[inst.flag])

    else:
      return 'set%s%s' % (InstDescriptor_SET.FLAGS[inst.flag], 'e' if inst.value == 1 else '')

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value = 1 if InstDescriptor_COND.evaluate(core, inst) is True else 0
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_CMP(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'cmp'
  opcode = DuckyOpcodes.CMP

  @staticmethod
  def execute(core, inst):
    core.CMP(core.registers.map[inst.reg].value, core.RI_VAL(inst))

class Inst_CMPU(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'cmpu'
  opcode = DuckyOpcodes.CMPU

  @staticmethod
  def execute(core, inst):
    core.CMP(core.registers.map[inst.reg].value, core.RI_VAL(inst), signed = False)

class Inst_J(InstDescriptor_Generic_Unary_RI):
  mnemonic = 'j'
  opcode   = DuckyOpcodes.J
  relative_address = True

  @staticmethod
  def execute(core, inst):
    core.JUMP(inst)

class Inst_BE(InstDescriptor_BRANCH):
  mnemonic = 'be'
  opcode   = DuckyOpcodes.BE

class Inst_BNE(InstDescriptor_BRANCH):
  mnemonic = 'bne'
  opcode   = DuckyOpcodes.BNE

class Inst_BNS(InstDescriptor_BRANCH):
  mnemonic = 'bns'
  opcode   = DuckyOpcodes.BNS

class Inst_BNZ(InstDescriptor_BRANCH):
  mnemonic = 'bnz'
  opcode   = DuckyOpcodes.BNZ

class Inst_BS(InstDescriptor_BRANCH):
  mnemonic = 'bs'
  opcode   = DuckyOpcodes.BS

class Inst_BZ(InstDescriptor_BRANCH):
  mnemonic = 'bz'
  opcode   = DuckyOpcodes.BZ

class Inst_BO(InstDescriptor_BRANCH):
  mnemonic = 'bo'
  opcode   = DuckyOpcodes.BO

class Inst_BNO(InstDescriptor_BRANCH):
  mnemonic = 'bno'
  opcode   = DuckyOpcodes.BNO

class Inst_BG(InstDescriptor_BRANCH):
  mnemonic = 'bg'
  opcode = DuckyOpcodes.BG

class Inst_BGE(InstDescriptor_BRANCH):
  mnemonic = 'bge'
  opcode = DuckyOpcodes.BGE

class Inst_BL(InstDescriptor_BRANCH):
  mnemonic = 'bl'
  opcode = DuckyOpcodes.BL

class Inst_BLE(InstDescriptor_BRANCH):
  mnemonic = 'ble'
  opcode = DuckyOpcodes.BLE

class Inst_SETE(InstDescriptor_SET):
  mnemonic = 'sete'
  opcode = DuckyOpcodes.SETE

class Inst_SETNE(InstDescriptor_SET):
  mnemonic = 'setne'
  opcode = DuckyOpcodes.SETNE

class Inst_SETZ(InstDescriptor_SET):
  mnemonic = 'setz'
  opcode = DuckyOpcodes.SETZ

class Inst_SETNZ(InstDescriptor_SET):
  mnemonic = 'setnz'
  opcode = DuckyOpcodes.SETNZ

class Inst_SETO(InstDescriptor_SET):
  mnemonic = 'seto'
  opcode = DuckyOpcodes.SETO

class Inst_SETNO(InstDescriptor_SET):
  mnemonic = 'setno'
  opcode = DuckyOpcodes.SETNO

class Inst_SETS(InstDescriptor_SET):
  mnemonic = 'sets'
  opcode = DuckyOpcodes.SETS

class Inst_SETNS(InstDescriptor_SET):
  mnemonic = 'setns'
  opcode = DuckyOpcodes.SETNS

class Inst_SETG(InstDescriptor_SET):
  mnemonic = 'setg'
  opcode = DuckyOpcodes.SETG

class Inst_SETGE(InstDescriptor_SET):
  mnemonic = 'setge'
  opcode = DuckyOpcodes.SETGE

class Inst_SETL(InstDescriptor_SET):
  mnemonic = 'setl'
  opcode = DuckyOpcodes.SETL

class Inst_SETLE(InstDescriptor_SET):
  mnemonic = 'setle'
  opcode = DuckyOpcodes.SETLE

#
# IO
#
class Inst_IN(InstDescriptor_Generic_Binary_R_RI):
  mnemonic      = 'in'
  opcode        = DuckyOpcodes.IN

  @staticmethod
  def execute(core, inst):
    port = core.RI_VAL(inst)
    core.check_protected_port(port)
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value = core.cpu.machine.ports[port].read_u16(port)

class Inst_INB(InstDescriptor_Generic_Binary_R_RI):
  mnemonic      = 'inb'
  opcode = DuckyOpcodes.INB

  @staticmethod
  def execute(core, inst):
    port = core.RI_VAL(inst)
    core.check_protected_port(port)
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value = core.cpu.machine.ports[port].read_u8(port)
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_OUT(InstDescriptor_Generic_Binary_RI_R):
  mnemonic      = 'out'
  opcode        = DuckyOpcodes.OUT

  @staticmethod
  def execute(core, inst):
    port = core.RI_VAL(inst)
    core.check_protected_port(port)
    core.cpu.machine.ports[port].write_u16(port, core.registers.map[inst.reg].value)

class Inst_OUTB(InstDescriptor_Generic_Binary_RI_R):
  mnemonic      = 'outb'
  opcode = DuckyOpcodes.OUTB

  @staticmethod
  def execute(core, inst):
    port = core.RI_VAL(inst)
    core.check_protected_port(port)
    core.cpu.machine.ports[port].write_u8(port, core.registers.map[inst.reg].value & 0xFF)

#
# Bit operations
#
class Inst_AND(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'and'
  opcode = DuckyOpcodes.AND

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value &= core.RI_VAL(inst)
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_OR(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'or'
  opcode = DuckyOpcodes.OR

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value |= core.RI_VAL(inst)
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_XOR(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'xor'
  opcode = DuckyOpcodes.XOR

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value ^= core.RI_VAL(inst)
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_NOT(InstDescriptor_Generic_Unary_R):
  mnemonic = 'not'
  opcode = DuckyOpcodes.NOT

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value = ~core.registers.map[inst.reg].value
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_SHIFTL(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'shiftl'
  opcode = DuckyOpcodes.SHIFTL

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value <<= core.RI_VAL(inst)
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_SHIFTR(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'shiftr'
  opcode = DuckyOpcodes.SHIFTR

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value >>= core.RI_VAL(inst)
    core.update_arith_flags(core.registers.map[inst.reg])

#
# Memory load/store operations
#
class Inst_CAS(InstDescriptor):
  mnemonic = 'cas'
  opcode = DuckyOpcodes.CAS

  operands = 'r,r,r'
  binary_format = [BF_REG('r_addr'), BF_REG('r_test'), BF_REG('r_rep')]

  @staticmethod
  def assemble_operands(logger, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    inst.r_addr = operands['register_n0']
    inst.r_test = operands['register_n1']
    inst.r_rep = operands['register_n2']

  @staticmethod
  def disassemble_operands(inst):
    return [
      REGISTER_NAMES[inst.r_addr],
      REGISTER_NAMES[inst.r_test],
      REGISTER_NAMES[inst.r_rep]
    ]

  @staticmethod
  def execute(core, inst):
    core.registers.flags.e = 0

    addr = core.DS_ADDR(core.registers.map[inst.r_addr].value)
    value = core.MEM_IN16(addr)

    if value == core.registers.map[inst.r_test].value:
      core.MEM_OUT16(addr, core.registers.map[inst.r_rep].value)
      core.registers.flags.e = 1

    else:
      core.registers.map[inst.r_test].value = value

class Inst_LW(InstDescriptor_Generic_Binary_R_A):
  mnemonic = 'lw'
  opcode = DuckyOpcodes.LW

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value = core.MEM_IN16(core.OFFSET_ADDR(inst))
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_LB(InstDescriptor_Generic_Binary_R_A):
  mnemonic = 'lb'
  opcode = DuckyOpcodes.LB

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value = core.MEM_IN8(core.OFFSET_ADDR(inst))
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_LI(InstDescriptor_Generic_Binary_R_I):
  mnemonic    = 'li'
  opcode = DuckyOpcodes.LI

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    core.registers.map[inst.reg].value = inst.immediate
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_STW(InstDescriptor_Generic_Binary_A_R):
  mnemonic    = 'stw'
  opcode = DuckyOpcodes.STW

  @staticmethod
  def execute(core, inst):
    core.MEM_OUT16(core.OFFSET_ADDR(inst), core.registers.map[inst.reg].value)

class Inst_STB(InstDescriptor_Generic_Binary_A_R):
  mnemonic    = 'stb'
  opcode = DuckyOpcodes.STB

  @staticmethod
  def execute(core, inst):
    core.MEM_OUT8(core.OFFSET_ADDR(inst), core.registers.map[inst.reg].value & 0xFF)

class Inst_MOV(InstDescriptor_Generic_Binary_R_R):
  mnemonic = 'mov'
  opcode = DuckyOpcodes.MOV

  @staticmethod
  def execute(core, inst):
    core.registers.map[inst.reg1].value = core.registers.map[inst.reg2].value

class Inst_SWP(InstDescriptor_Generic_Binary_R_R):
  mnemonic = 'swp'
  opcode = DuckyOpcodes.SWP

  @staticmethod
  def execute(core, inst):
    v = core.registers.map[inst.reg1].value
    core.registers.map[inst.reg1].value = core.registers.map[inst.reg2].value
    core.registers.map[inst.reg2].value = v

class Inst_MUL(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'mul'
  opcode = DuckyOpcodes.MUL

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    r = core.registers.map[inst.reg]
    x = i16(r.value).value
    y = i16(core.RI_VAL(inst)).value
    r.value = x * y
    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_DIV(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'div'
  opcode = DuckyOpcodes.DIV

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    r = core.registers.map[inst.reg]
    x = i16(r.value).value
    y = i16(core.RI_VAL(inst)).value

    if abs(y) > abs(x):
      r.value = 0

    else:
      r.value = x // y

    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_UDIV(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'udiv'
  opcode = DuckyOpcodes.UDIV

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    r = core.registers.map[inst.reg]
    x = u16(r.value).value
    y = u16(core.RI_VAL(inst)).value

    r.value = x // y

    core.update_arith_flags(core.registers.map[inst.reg])

class Inst_MOD(InstDescriptor_Generic_Binary_R_RI):
  mnemonic = 'mod'
  opcode = DuckyOpcodes.MOD

  @staticmethod
  def execute(core, inst):
    core.check_protected_reg(inst.reg)
    r = core.registers.map[inst.reg]
    x = i16(r.value).value
    y = i16(core.RI_VAL(inst)).value
    r.value = x % y
    core.update_arith_flags(core.registers.map[inst.reg])

class DuckyInstructionSet(InstructionSet):
  instruction_set_id = 0
  opcodes = DuckyOpcodes

Inst_NOP(DuckyInstructionSet)
Inst_INT(DuckyInstructionSet)
Inst_RETINT(DuckyInstructionSet)
Inst_CALL(DuckyInstructionSet)
Inst_RET(DuckyInstructionSet)
Inst_CLI(DuckyInstructionSet)
Inst_STI(DuckyInstructionSet)
Inst_HLT(DuckyInstructionSet)
Inst_RST(DuckyInstructionSet)
Inst_IDLE(DuckyInstructionSet)
Inst_PUSH(DuckyInstructionSet)
Inst_POP(DuckyInstructionSet)
Inst_INC(DuckyInstructionSet)
Inst_DEC(DuckyInstructionSet)
Inst_ADD(DuckyInstructionSet)
Inst_SUB(DuckyInstructionSet)
Inst_CMP(DuckyInstructionSet)
Inst_J(DuckyInstructionSet)
Inst_IN(DuckyInstructionSet)
Inst_INB(DuckyInstructionSet)
Inst_OUT(DuckyInstructionSet)
Inst_OUTB(DuckyInstructionSet)
Inst_AND(DuckyInstructionSet)
Inst_OR(DuckyInstructionSet)
Inst_XOR(DuckyInstructionSet)
Inst_NOT(DuckyInstructionSet)
Inst_SHIFTL(DuckyInstructionSet)
Inst_SHIFTR(DuckyInstructionSet)
Inst_LW(DuckyInstructionSet)
Inst_LB(DuckyInstructionSet)
Inst_LI(DuckyInstructionSet)
Inst_STW(DuckyInstructionSet)
Inst_STB(DuckyInstructionSet)
Inst_MOV(DuckyInstructionSet)
Inst_SWP(DuckyInstructionSet)
Inst_MUL(DuckyInstructionSet)
Inst_DIV(DuckyInstructionSet)
Inst_UDIV(DuckyInstructionSet)
Inst_MOD(DuckyInstructionSet)
Inst_CMPU(DuckyInstructionSet)
Inst_CAS(DuckyInstructionSet)
Inst_SIS(DuckyInstructionSet)

# Branching instructions
Inst_BE(DuckyInstructionSet)
Inst_BNE(DuckyInstructionSet)
Inst_BZ(DuckyInstructionSet)
Inst_BNZ(DuckyInstructionSet)
Inst_BO(DuckyInstructionSet)
Inst_BNO(DuckyInstructionSet)
Inst_BS(DuckyInstructionSet)
Inst_BNS(DuckyInstructionSet)
Inst_BG(DuckyInstructionSet)
Inst_BGE(DuckyInstructionSet)
Inst_BL(DuckyInstructionSet)
Inst_BLE(DuckyInstructionSet)

# SET* instructions
Inst_SETE(DuckyInstructionSet)
Inst_SETNE(DuckyInstructionSet)
Inst_SETZ(DuckyInstructionSet)
Inst_SETNZ(DuckyInstructionSet)
Inst_SETO(DuckyInstructionSet)
Inst_SETNO(DuckyInstructionSet)
Inst_SETS(DuckyInstructionSet)
Inst_SETNS(DuckyInstructionSet)
Inst_SETG(DuckyInstructionSet)
Inst_SETGE(DuckyInstructionSet)
Inst_SETL(DuckyInstructionSet)
Inst_SETLE(DuckyInstructionSet)

DuckyInstructionSet.init()

INSTRUCTION_SETS = {
  DuckyInstructionSet.instruction_set_id: DuckyInstructionSet
}

def get_instruction_set(i, exc = None):
  from ..cpu import InvalidInstructionSetError
  exc = exc or InvalidInstructionSetError

  if i not in INSTRUCTION_SETS:
    raise exc(i)

  return INSTRUCTION_SETS[i]



import mmap

from six import iteritems, iterkeys, itervalues, PY2
from six.moves import range

from ctypes import LittleEndianStructure, c_ubyte, c_ushort, c_uint

from ..interfaces import ISnapshotable
from ..errors import AccessViolationError, InvalidResourceError
from ..util import align, sizeof_fmt
from ..snapshot import SnapshotNode

import enum

# Types
from ctypes import c_byte as i8  # NOQA
from ctypes import c_short as i16  # NOQA
from ctypes import c_int as i32  # NOQA

from ctypes import c_ubyte as u8  # NOQA
from ctypes import c_ushort as u16  # NOQA
from ctypes import c_uint as u32  # NOQA

PAGE_SHIFT = 8
#: Size of memory page, in bytes.
PAGE_SIZE = (1 << PAGE_SHIFT)
PAGE_MASK = (~(PAGE_SIZE - 1))

SEGMENT_SHIFT = 16
#: Size of segment, in pages
SEGMENT_SIZE  = 256  # pages
SEGMENT_PROTECTED = 0  # first segment is already allocated

class MMOperationList(enum.IntEnum):
  MPROTECT = 1
  MTELL    = 2
  ALLOC    = 3
  FREE     = 4
  UNUSED   = 5
  MMAP     = 6
  UNMMAP   = 7

MM_FLAG_READ    = 0x0001
MM_FLAG_WRITE   = 0x0002
MM_FLAG_EXECUTE = 0x0004
MM_FLAG_DIRTY   = 0x0008

MM_FLAG_CS      = 0x1000

from ..util import UINT8_FMT, UINT16_FMT, UINT24_FMT, UINT32_FMT, ADDR_FMT  # noqa

def SEGM_FMT(segment):
  return UINT8_FMT(segment)

def SIZE_FMT(size):
  return str(size)

def OFFSET_FMT(offset):
  s = '-' if offset < 0 else ''

  return '{}0x{:04X}'.format(s, abs(offset))

class MalformedBinaryError(Exception):
  pass

class UInt8(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('u8', c_ubyte)
  ]

  def __repr__(self):
    return '<(u8) 0x{:02X}>'.format(self.u8)

class UInt16(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('u16', c_ushort)
  ]

  def __repr__(self):
    return '<(u16) 0x{:04X}>'.format(self.u16)

# Yes, this one is larger but it's used only for transporting
# addresses between CPUs and memory controller => segment
# register and u16 have to fit in.
class UInt24(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('u24', c_uint, 24)
  ]

  def __repr__(self):
    return '<(u24) 0x{:06X}>'.format(self.u24)

class UInt32(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('u32', c_uint)
  ]

  def __repr__(self):
    return '<(u32) 0x{:06X}>'.format(self.u32)

def segment_base_addr(segment):
  return segment * SEGMENT_SIZE * PAGE_SIZE

def segment_addr_to_addr(segment, addr):
  return segment * SEGMENT_SIZE * PAGE_SIZE + addr

def addr_to_segment(addr):
  return (addr & 0xFF0000) >> 16

def addr_to_page(addr):
  return (addr & PAGE_MASK) >> PAGE_SHIFT

def addr_to_offset(addr):
  return addr & (PAGE_SIZE - 1)

def area_to_pages(addr, size):
  return ((addr & PAGE_MASK) >> PAGE_SHIFT, align(PAGE_SIZE, size) // PAGE_SIZE)

class MemoryPageState(SnapshotNode):
  def __init__(self, *args, **kwargs):
    super(MemoryPageState, self).__init__('index', 'read', 'write', 'execute', 'dirty', 'stack', 'cache', 'content')

class MemoryPage(object):
  """
  Base class for all memory pages of any kinds.

  Memory page has a set of boolean flags that determine access to and behavior
  of the page.

  +-------------+-----------------------------------------------------------------------------+-----------+
  | Flag        | Meaning                                                                     | Default   |
  +-------------+-----------------------------------------------------------------------------+-----------+
  | ``read``    | page is readable by executed instructions                                   | ``False`` |
  +-------------+-----------------------------------------------------------------------------+-----------+
  | ``write``   | page is writable by executed instructions                                   | ``False`` |
  +-------------+-----------------------------------------------------------------------------+-----------+
  | ``execute`` | content of the page can be used as executable instructions                  | ``False`` |
  +-------------+-----------------------------------------------------------------------------+-----------+
  | ``dirty``   | there have been write access to this page, its content has changed          | ``False`` |
  +-------------+-----------------------------------------------------------------------------+-----------+
  | ``stack``   | page is a stack                                                             | ``False`` |
  +-------------+-----------------------------------------------------------------------------+-----------+
  | ``cache``   | all RW accesses to this page made by CPU cores go through their data caches | ``True``  |
  +-------------+-----------------------------------------------------------------------------+-----------+

  :param ducky.mm.MemoryController controller: Controller that owns this page.
  :param int index: Serial number of this page.
  """

  def __init__(self, controller, index):
    super(MemoryPage, self).__init__()

    self.controller = controller
    self.index = index

    self.DEBUG = self.controller.DEBUG
    self.INFO = self.controller.INFO
    self.WARN = self.controller.WARN
    self.ERROR = self.controller.ERROR
    self.EXCEPTION = self.controller.EXCEPTION

    self.base_address = self.index * PAGE_SIZE
    self.segment_address = self.base_address % (SEGMENT_SIZE * PAGE_SIZE)

    self.read    = False
    self.write   = False
    self.execute = False
    self.dirty   = False
    self.stack   = False
    self.cache   = False

  def __repr__(self):
    return '<%s index=%i, base=%s, segment_addr=%s, flags=%s>' % (self.__class__.__name__, self.index, ADDR_FMT(self.base_address), ADDR_FMT(self.segment_address), self.flags_str())

  def save_state(self, parent):
    """
    Create state of this page, and attach it to snapshot tree.

    :param parent: Parent snapshot node.
    :type parent: ducky.snapshot.SnapshotNode
    """

    state = parent.add_child('page_{}'.format(self.index), MemoryPageState())

    state.index = self.index

    state.content = [ord(i) if isinstance(i, str) else i for i in self.data]
    state.read = self.read
    state.write = self.write
    state.execute = self.execute
    state.dirty = self.dirty
    state.stack = self.stack
    state.cache = self.cache

    return state

  def load_state(self, state):
    """
    Restore page from a snapshot.
    """

    for i in range(0, PAGE_SIZE):
      self.data[i] = state.content[i]

    self.read = state.read
    self.write = state.write
    self.execute = state.execute
    self.dirty = state.dirty
    self.stack = state.stack
    self.cache = state.cache

  def flags_reset(self):
    """
    Reset all page flags to ``False``.
    """

    self.read = False
    self.write = False
    self.execute = False
    self.dirty = False
    self.cache = True

  def flags_str(self):
    """
    Return string representing page flags. Each position in string represents
    one of the flags. If flag is ``False``, ``-`` is emmited, otherwise a letter
    symbolizing that flag stands on the position.

    E.g. ``RW-D`` are flags of page that is readable, writable and dirty but
    not executable.

    :rtype: string
    """

    return ''.join([
      'R' if self.read else '-',
      'W' if self.write else '-',
      'X' if self.execute else '-',
      'D' if self.dirty else '-',
      'C' if self.cache else '-'
    ])

  def check_access(self, offset, access):
    """
    Return ``True`` if required access is valid for this page.

    :param int offset: Offset of the byte caller wants to manipulate. This
      parameter is only informative, used for debugging purposes.
    :param string access: One of ``read``, ``write`` or ``execute``.
    :rtype: bool
    :raises ducky.errors.AccessViolationError: when access is denied.
    """

    self.DEBUG('mp.check_access: page=%s, offset=%s, access=%s, %s', self.index, offset, access, self.flags_str())

    if access == 'read' and not self.read:
      raise AccessViolationError('Not allowed to read from memory: page={}, offset={}, addr={}'.format(self.index, offset, ADDR_FMT(self.base_address + offset)))

    if access == 'write' and not self.write:
      raise AccessViolationError('Not allowed to write to memory: page={}, offset={}, addr={}'.format(self.index, offset, ADDR_FMT(self.base_address + offset)))

    if access == 'execute' and not self.execute:
      raise AccessViolationError('Not allowed to execute from memory: page={}, offset={}, addr={}'.format(self.index, offset, ADDR_FMT(self.base_address + offset)))

    return True

  def __len__(self):
    """
    :return: length of this page. By default, all pages have the same length.
    :rtype: int
    """

    return PAGE_SIZE

  def do_clear(self):
    """
    Clear page.

    This operation is implemented by child classes.
    """

    raise AccessViolationError('Not allowed to clear memory on this address: page={}'.format(self.index))

  def do_read_u8(self, offset):
    """
    Read byte.

    This operation is implemented by child classes.

    :param int offset: offset of requested byte.
    :rtype: int
    """

    raise AccessViolationError('Not allowed to access memory on this address: page=%s, offset={}'.format(self.index, offset))

  def do_read_u16(self, offset):
    """
    Read word.

    This operation is implemented by child classes.

    :param int offset: offset of requested word.
    :rtype: int
    """

    raise AccessViolationError('Not allowed to access memory on this address: page=%s, offset={}'.format(self.index, offset))

  def do_read_u32(self, offset):
    """
    Read longword.

    This operation is implemented by child classes.

    :param int offset: offset of requested longword.
    :rtype: int
    """

    raise AccessViolationError('Not allowed to access memory on this address: page=%s, offset={}'.format(self.index, offset))

  def do_write_u8(self, offset, value):
    """
    Write byte.

    This operation is implemented by child classes.

    :param int offset: offset of requested byte.
    :param int value: value to write into memory.
    """

    raise AccessViolationError('Not allowed to access memory on this address: page=%s, offset={}'.format(self.index, offset))

  def do_write_u16(self, offset, value):
    """
    Write word.

    This operation is implemented by child classes.

    :param int offset: offset of requested word.
    :param int value: value to write into memory.
    """

    raise AccessViolationError('Not allowed to access memory on this address: page=%s, offset={}'.format(self.index, offset))

  def do_write_u32(self, offset, value):
    """
    Write longword.

    This operation is implemented by child classes.

    :param int offset: offset of requested longword.
    :param int value: value to write into memory.
    """

    raise AccessViolationError('Not allowed to access memory on this address: page=%s, offset={}'.format(self.index, offset))

  def clear(self, privileged = False):
    """
    Clear page. Checks page flags, and calls ``do_clear`` method to perform the
    `clear` operation.

    :param bool privileged: if not ``True``, page access is checked first.
    :raises ducky.errors.AccessViolationError: when page is not writtable.
    """

    self.DEBUG('mp.clear: page=%s, priv=%s', self.index, privileged)

    if privileged is not True:
      self.check_access(self.base_address, 'write')

    self.do_clear()

  def read_u8(self, offset, privileged = False):
    """
    Read byte. Checks page flags, and calls ``do_read_u8`` method to perform
    the `read` operation.

    :param int offset: offset of requested byte.
    :param bool privileged: if not ``True``, page access is checked first.
    :rtype: int
    :raises ducky.errors.AccessViolationError: when page is not readable.
    """

    self.DEBUG('mp.read_u8: page=%s, offset=%s, priv=%s', self.index, offset, privileged)

    if privileged is not True:
      self.check_access(offset, 'read')

    return self.do_read_u8(offset)

  def read_u16(self, offset, privileged = False):
    """
    Read word. Checks page flags, and calls ``do_read_u16`` method to perform
    the `read` operation.

    :param int offset: offset of requested word.
    :param bool privileged: if not ``True``, page access is checked first.
    :raises ducky.errors.AccessViolationError: when page is not readable.
    """

    self.DEBUG('mp.read_u16: page=%s, offset=%s, priv=%s', self.index, offset, privileged)

    if privileged is not True:
      self.check_access(offset, 'read')

    return self.do_read_u16(offset)

  def read_u32(self, offset, privileged = False, not_execute = True):
    """
    Read longword. Checks page flags, and calls ``do_read_u32`` method to
    perform the `read` operation.

    :param int offset: offset of requested longword.
    :param bool privileged: if not ``True``, page access is checked first.
    :param bool execute: if ``False``, read value will be interpreted as
      an instruction and executed
    :raises ducky.errors.AccessViolationError: when page is not readable.
    """

    self.DEBUG('mp.read_u32: page=%s, offset=%s, priv=%s, not_execute=%s', self.index, offset, privileged, not_execute)

    if privileged is not True:
      self.check_access(offset, 'read')

    if not_execute is not True:
      self.check_access(offset, 'execute')

    return self.do_read_u32(offset)

  def write_u8(self, offset, value, privileged = False, dirty = True):
    """
    Write byte. Checks page flags, and calls ``do_write_u8`` method to perform
    the `write` operation.

    :param int offset: offset of requested byte.
    :param int value: value to write into memory.
    :param bool privileged: if not ``True``, page access is checked first.
    :param bool dirty: if ``True``, page ``dirty`` flag is set to ``True``.
    :raises ducky.errors.AccessViolationError: when page is not writable.
    """

    self.DEBUG('mp.write_u8: page=%s, offset=%s, value=%s, priv=%s, dirty=%s', self.index, offset, value, privileged, dirty)

    if privileged is not True:
      self.check_access(offset, 'write')

    self.do_write_u8(offset, value)
    if dirty:
      self.dirty = True

  def write_u16(self, offset, value, privileged = False, dirty = True):
    """
    Write word. Checks page flags, and calls ``do_write_u16`` method to perform
    the `write` operation.

    :param int offset: offset of requested word.
    :param int value: value to write into memory.
    :param bool privileged: if not ``True``, page access is checked first.
    :param bool dirty: if ``True``, page ``dirty`` flag is set to ``True``.
    :raises ducky.errors.AccessViolationError: when page is not writable.
    """

    self.DEBUG('mp.write_u16: page=%s, offset=%s, value=%s, priv=%s, dirty=%s', self.index, offset, value, privileged, dirty)

    if privileged is not True:
      self.check_access(offset, 'write')

    self.do_write_u16(offset, value)
    if dirty:
      self.dirty = True

  def write_u32(self, offset, value, privileged = False, dirty = True):
    """
    Write longbyte. Checks page flags, and calls ``do_write_u32`` method to
    perform the `write` operation.

    :param int offset: offset of requested longword.
    :param int value: value to write into memory.
    :param bool privileged: if not ``True``, page access is checked first.
    :param bool dirty: if ``True``, page ``dirty`` flag is set to ``True``.
    :raises ducky.errors.AccessViolationError: when page is not writable.
    """

    self.DEBUG('mp.write_u32: page=%s, offset=%s, value=%s, priv=%s, dirty=%s', self.index, offset, value, privileged, dirty)

    if privileged is not True:
      self.check_access(offset, 'write')

    self.do_write_u32(offset, value)
    if dirty:
      self.dirty = True

class AnonymousMemoryPage(MemoryPage):
  """
  "Anonymous" memory page - this page is just a plain array of bytes, and is
  not backed by any storage. Its content lives only in the memory.

  Page is created with all bytes set to zero.
  """

  def __init__(self, controller, index):
    super(AnonymousMemoryPage, self).__init__(controller, index)

    self.data = bytearray([0 for _ in range(0, PAGE_SIZE)])

  def do_clear(self):
    for i in range(0, PAGE_SIZE):
      self.data[i] = 0

  def do_read_u8(self, offset):
    self.DEBUG('mp.do_read_u8: page=%s, offset=%s', self.index, offset)

    return self.data[offset]

  def do_read_u16(self, offset):
    self.DEBUG('mp.do_read_u16: page=%s, offset=%s', self.index, offset)

    return self.data[offset] | (self.data[offset + 1] << 8)

  def do_read_u32(self, offset):
    self.DEBUG('mp.do_read_u32: page=%s, offset=%s', self.index, offset)

    return self.data[offset] | (self.data[offset + 1] << 8) | (self.data[offset + 2] << 16) | (self.data[offset + 3] << 24)

  def do_write_u8(self, offset, value):
    self.DEBUG('mp.do_write_u8: page=%s, offset=%s, value=%s', self.index, offset, value)

    self.data[offset] = value

  def do_write_u16(self, offset, value):
    self.DEBUG('mp.do_write_u16: page=%s, offset=%s, value=%s', self.index, offset, value)

    self.data[offset]     =  value & 0x00FF
    self.data[offset + 1] = (value & 0xFF00) >> 8

  def do_write_u32(self, offset, value):
    self.DEBUG('mp.do_write_u32: page=%s, offset=%s, value=%s', self.index, offset, value)

    self.data[offset]     =  value &       0xFF
    self.data[offset + 1] = (value &     0xFF00) >> 8
    self.data[offset + 2] = (value &   0xFF0000) >> 16
    self.data[offset + 3] = (value & 0xFF000000) >> 24

class ExternalMemoryPage(MemoryPage):
  """
  Memory page backed by an external source. Source is an array of bytes,
  and can be provided by device driver, mmaped file, or by any other mean.
  """

  def __init__(self, controller, index, data, offset = 0):
    super(ExternalMemoryPage, self).__init__(controller, index)

    self.data = data
    self.offset = offset

  def __repr__(self):
    return '<%s index=%i, base=%s, segment_addr=%s, flags=%s, offset=%s>' % (self.__class__.__name__, self.index, ADDR_FMT(self.base_address), ADDR_FMT(self.segment_address), self.flags_str(), ADDR_FMT(self.offset))

  def save_state(self, parent):
    state = super(ExternalMemoryPage, self).save_state(parent)

    state.content = [ord(i) if isinstance(i, str) else i for i in self.data[self.offset:self.offset + PAGE_SIZE]]

  def do_clear(self):
    for i in range(0, PAGE_SIZE):
      self.data[i] = 0

  def get(self, offset):
    """
    Get one byte from page. Override this method in case you need a different
    offset of requested byte.

    :param int offset: offset of the requested byte.
    :rtype: int
    :returns: byte at position in page.
    """

    return self.data[self.offset + offset]

  def put(self, offset, b):
    """
    Put one byte into page. Override this method in case you need a different
    offset of requested byte.

    :param int offset: offset of modified byte.
    :param int b: new value.
    """

    self.data[self.offset + offset] = b

  def do_read_u8(self, offset):
    self.DEBUG('%s.do_read_u8: page=%s, offset=%s', self.__class__.__name__, self.index, offset)

    return self.get(offset)

  def do_read_u16(self, offset):
    self.DEBUG('%s.do_read_u16: page=%s, offset=%s', self.__class__.__name__, self.index, offset)

    return self.get(offset) | (self.get(offset + 1) << 8)

  def do_read_u32(self, offset):
    self.DEBUG('%s.do_read_u32: page=%s, offset=%s', self.__class__.__name__, self.index, offset)

    return self.get(offset) | (self.get(offset + 1) << 8) | (self.get(offset + 2) << 16) | (self.get(offset + 3) << 24)

  def do_write_u8(self, offset, value):
    self.DEBUG('%s.do_write_u8: page=%s, offset=%s, value=%s', self.__class__.__name__, self.index, offset, value)

    self.put(offset, value)

  def do_write_u16(self, offset, value):
    self.DEBUG('%s.do_write_u16: page=%s, offset=%s, value=%s', self.__class__.__name__, self.index, offset, value)

    self.put(offset, value & 0x00FF)
    self.put(offset + 1, (value & 0xFF00) >> 8)

  def do_write_u32(self, offset, value):
    self.DEBUG('%s.do_write_u32: page=%s, offset=%s, value=%s', self.__class__.__name__, self.index, offset, value)

    self.put(offset, value & 0x00FF)
    self.put(offset + 1, (value & 0xFF00) >> 8)
    self.put(offset + 2, (value & 0xFF0000) >> 16)
    self.put(offset + 3, (value & 0xFF000000) >> 24)

class MMapMemoryPage(ExternalMemoryPage):
  """
  Memory page backed by an external file that is accessible via ``mmap()``
  call. It's a part of one of mm.MMapArea instances, and if such area was
  opened as `shared`, every change in this page content will affect the
  content of external file, and vice versa, every change of external file
  will be reflected in content of this page (if this page lies in affected
  area).
  """

  def __init__(self, area, *args, **kwargs):
    super(MMapMemoryPage, self).__init__(*args, **kwargs)

    self.area = area

    if PY2:
      self.get, self.put = self._get_py2, self._put_py2

    else:
      self.get, self.put = self._get_py3, self._put_py3

  def _get_py2(self, offset):
    return ord(self.data[self.offset + offset])

  def _put_py2(self, offset, b):
    self.data[self.offset + offset] = chr(b)

  def _get_py3(self, offset):
    return self.data[self.offset + offset]

  def _put_py3(self, offset, b):
    self.data[self.offset + offset] = b

class MMapAreaState(SnapshotNode):
  def __init__(self):
    super(MMapAreaState, self).__init__('address', 'size', 'path', 'offset')

class MMapArea(object):
  def __init__(self, ptr, address, size, file_path, offset, pages_start, pages_cnt):
    super(MMapArea, self).__init__()

    self.ptr = ptr
    self.address = address
    self.size = size
    self.file_path = file_path
    self.offset = offset
    self.pages_start = pages_start
    self.pages_cnt = pages_cnt

  def __repr__(self):
    return '<MMapArea: address=%s, size=%s, filepath=%s, pages-start=%s, pages-cnt=%i>' % (ADDR_FMT(self.address), self.size, self.file_path, self.pages_start, self.pages_cnt)

  def save_state(self, parent):
    pass

  def load_state(self, state):
    pass

class MemoryRegionState(SnapshotNode):
  def __init__(self):
    super(MemoryRegionState, self).__init__('name', 'address', 'size', 'flags', 'pages_start', 'pages_cnt')

class MemoryRegion(ISnapshotable, object):
  region_id = 0

  def __init__(self, mc, name, address, size, flags):
    super(MemoryRegion, self).__init__()

    self.memory = mc

    self.id = MemoryRegion.region_id
    MemoryRegion.region_id += 1

    self.name = name
    self.address = address
    self.size = size
    self.flags = flags

    self.pages_start, self.pages_cnt = area_to_pages(self.address, self.size)

    self.memory.machine.DEBUG('MemoryRegion: name=%s, address=%s, size=%s, flags=%s, pages_start=%s, pages_cnt=%s', name, address, size, self.flags.to_string(), self.pages_start, self.pages_cnt)

  def save_state(self, parent):
    state = parent.add_child('memory_region_{}'.format(self.id), MemoryRegionState())

    state.name = self.name
    state.address = self.address
    state.size = self.size
    state.flags = self.flags.to_uint16()
    state.pages_start = self.pages_start
    state.pages_cnt = self.pages_cnt

  def load_state(self, state):
    pass

class MemoryState(SnapshotNode):
  def __init__(self):
    super(MemoryState, self).__init__('size', 'segments')

  def get_page_states(self):
    return [__state for __name, __state in iteritems(self.get_children()) if __name.startswith('page_')]

class MemoryController(object):
  """
  Memory controller handles all operations regarding main memory.

  :param ducky.machine.Machine machine: virtual machine that owns this controller.
  :param int size: size of memory, in bytes.
  :raises ducky.errors.InvalidResourceError: when memory size is not multiple of
    :py:data:`ducky.mm.PAGE_SIZE`, or when size is not multiply of
    :py:data:`ducky.mm.SEGMENT_SIZE` pages.
  """

  def __init__(self, machine, size = 0x1000000):
    if size % PAGE_SIZE != 0:
      raise InvalidResourceError('Memory size must be multiple of PAGE_SIZE')

    if size % (SEGMENT_SIZE * PAGE_SIZE) != 0:
      raise InvalidResourceError('Memory size must be multiple of SEGMENT_SIZE')

    self.machine = machine

    # Setup logging - create our local shortcuts to machine' logger
    self.DEBUG = self.machine.DEBUG
    self.INFO = self.machine.INFO
    self.WARN = self.machine.WARN
    self.ERROR = self.machine.ERROR
    self.EXCEPTION = self.machine.EXCEPTION

    self.force_aligned_access = self.machine.config.getbool('memory', 'force-aligned-access', default = False)

    self.__size = size
    self.__pages_cnt = size // PAGE_SIZE
    self.__pages = {}

    self.__segments_cnt = size // (SEGMENT_SIZE * PAGE_SIZE)
    self.__segments = {}

    # mmap
    self.opened_mmap_files = {}  # path: (cnt, file)
    self.mmap_areas = {}

  def save_state(self, parent):
    self.DEBUG('mc.save_state')

    state = parent.add_child('memory', MemoryState())

    state.size = self.__size

    state.segments = []
    for segment in iterkeys(self.__segments):
      state.segments.append(segment)

    for page in itervalues(self.__pages):
      page.save_state(state)

  def load_state(self, state):
    self.size = state.size

    for segment in state.segments:
      self.__segments[segment] = True

    for page_state in state.get_children():
      page = self.page(page_state.index)
      page.load_state(page_state)

  def alloc_segment(self):
    """
    Reserve one of free memory segments.

    :returns: index of reserved segment.
    :rtype: int
    :raises ducky.errors.InvalidResourceError: when there are no free segments.
    """

    self.DEBUG('mc.alloc_segment')

    for i in range(0, self.__segments_cnt):
      if i in self.__segments:
        continue

      # No SegmentMapEntry flags are in use right now, just keep this option open
      self.DEBUG('mc.alloc_segment: segment=%s', i)

      self.__segments[i] = True
      return i

    raise InvalidResourceError('No free segment available')

  def __set_page(self, pg):
    """
    Install page object for a specific memory page.

    :param ducky.mm.MemoryPage pg: page to be installed
    :returns: installed page
    :rtype: :py:class:`ducky.mm.MemoryPage`
    """

    assert pg.index not in self.__pages

    self.__pages[pg.index] = pg
    return pg

  def __remove_page(self, pg):
    """
    Removes page object for a specific memory page.

    :param ducky.mm.MemoryPage pg: page to be removed
    """

    assert pg.index in self.__pages

    del self.__pages[pg.index]

  def __alloc_page(self, index):
    """
    Allocate new anonymous page for usage. The first available index is used.

    Be aware that this method does NOT check if page is already allocated. If
    it is, it is just overwritten by new anonymous page.

    :param int index: index of requested page.
    :returns: newly reserved page.
    :rtype: :py:class:`ducky.mm.AnonymousMemoryPage`
    """

    return self.__set_page(AnonymousMemoryPage(self, index))

  def alloc_specific_page(self, index):
    """
    Allocate new anonymous page with specific index for usage.

    :param int index: allocate page with this particular index.
    :returns: newly reserved page.
    :rtype: :py:class:`ducky.mm.AnonymousMemoryPage`
    :raises ducky.errors.AccessViolationError: when page is already allocated.
    """

    self.DEBUG('mc.alloc_specific_page: index=%s', index)

    if index in self.__pages:
      raise AccessViolationError('Page {} is already allocated'.format(index))

    return self.__alloc_page(index)

  def alloc_pages(self, segment = None, count = 1):
    """
    Allocate continuous sequence of anonymous pages.

    :param int segment: if not ``None``, allocated pages will be from this
      segment.
    :param int count: number of requested pages.
    :returns: list of newly allocated pages.
    :rtype: ``list`` of :py:class:`ducky.mm.AnonymousMemoryPage`
    :raises ducky.errors.InvalidResourceError: when there is no available sequence of
      pages.
    """

    self.DEBUG('mc.alloc_pages: segment=%s, count=%s', segment if segment else '', count)

    if segment is not None:
      pages_start = segment * SEGMENT_SIZE
      pages_cnt = SEGMENT_SIZE
    else:
      pages_start = 0
      pages_cnt = self.__pages_cnt

    self.DEBUG('mc.alloc_pages: page=%s, cnt=%s', pages_start, pages_cnt)

    for i in range(pages_start, pages_start + pages_cnt):
      for j in range(i, i + count):
        if j in self.__pages:
          break

      else:
        return [self.__alloc_page(j) for j in range(i, i + count)]

    raise InvalidResourceError('No sequence of free pages available')

  def alloc_page(self, segment = None):
    """
    Allocate new anonymous page for usage. The first available index is used.

    :param int segment: if not ``None``, allocated page will be from this
      segment.
    :returns: newly reserved page.
    :rtype: :py:class:`ducky.mm.AnonymousMemoryPage`
    :raises ducky.errors.InvalidResourceError: when there is no available page.
    """

    self.DEBUG('mc.alloc_page: segment=%s', segment if segment else '')

    if segment is not None:
      pages_start = segment * SEGMENT_SIZE
      pages_cnt = SEGMENT_SIZE
    else:
      pages_start = 0
      pages_cnt = self.__pages_cnt

    self.DEBUG('mc.alloc_page: page=%s, cnt=%s', pages_start, pages_cnt)

    for i in range(pages_start, pages_start + pages_cnt):
      if i not in self.__pages:
        self.DEBUG('mc.alloc_page: page=%s', i)
        return self.__alloc_page(i)

    raise InvalidResourceError('No free page available')

  def register_page(self, pg):
    """
    Install page object for a specific memory page. This method is intended
    for external objects, e.g. device drivers to install their memory page
    objects to handle memory-mapped IO.

    :param ducky.mm.MemoryPage pg: page to be installed
    :returns: installed page
    :rtype: :py:class:`ducky.mm.AnonymousMemoryPage`
    :raises ducky.errors.AccessViolationError: when there is already allocated page
    """

    self.DEBUG('mc.register_page: pg=%s', pg)

    if pg.index in self.__pages:
      raise AccessViolationError('Page {} is already allocated'.format(pg.index))

    return self.__set_page(pg)

  def unregister_page(self, pg):
    """
    Remove page object for a specific memory page. This method is intende
    for external objects, e.g. device drivers to remove their memory page objects
    handling memory-mapped IO.

    :param ducky.mm.MemoryPage pg: page to be removed
    :raises ducky.errors.AccessViolationError: when there is no allocated page
    """

    self.DEBUG('mc.unregister_page: pg=%s', pg)

    if pg.index not in self.__pages:
      raise AccessViolationError('Page {} is not allocated'.format(pg.index))

    self.__remove_page(pg)

  def alloc_stack(self, segment = None):
    """
    Allocate page for a stack. Such page does not differ from other but this
    pages are requested from different places of virtual machine, therefore
    this shortcut method.

    :param int segment: if not ``None``, , allocated page will be from this
      segment.
    :returns: newly allocated page, and base address of the next memory page.
      This address is in fact a stack pointer for storing the first value on
      newly allocated stack (`decrement and store`).
    :rtype: (:py:class:`ducky.mm.AnonymousMemoryPage`, ``int``)
    :raises ducky.errors.InvalidResourceError: when there is no available page.
    """

    pg = self.alloc_page(segment)
    pg.flags_reset()
    pg.read = True
    pg.write = True
    pg.stack = True

    return (pg, pg.segment_address + PAGE_SIZE)

  def alloc_ivt(self, address):
    pg = self.__alloc_page(addr_to_page(address))
    pg.read = True
    pg.cache = False

    return pg

  def free_page(self, page):
    """
    Free memory page when it's no longer needed.

    :param ducky.mm.MemoryPage page: page to be freed.
    """

    self.DEBUG('mc.free_page: page=%i, base=%s, segment=%s', page.index, page.base_address, page.segment_address)

    self.__remove_page(page)

  def free_pages(self, page, count = 1):
    """
    Free a continuous sequence of pages when they are no longer needed.

    :param ducky.mm.MemoryPage page: first page in series.
    :param int count: number of pages.
    """

    self.DEBUG('mc.free_pages: page=%i, base=%s, segment=%s, count=%s', page.index, page.base_address, page.segment_address, count)

    for i in range(page.index, page.index + count):
      self.free_page(self.__pages[i])

  def page(self, index):
    """
    Return memory page, specified by its index from the beginning of memory.

    :param int index: index of requested page.
    :rtype: :py:class:`ducky.mm.MemoryPage`
    :raises ducky.errors.AccessViolationError: when requested page is not allocated.
    """

    if index not in self.__pages:
      raise AccessViolationError('Page {} not allocated yet'.format(index))

    return self.__pages[index]

  def pages(self, pages_start = 0, pages_cnt = None, ignore_missing = False):
    """
    Return list of memory pages.

    :param int pages_start: index of the first page, 0 by default.
    :param int pages_cnt: number of pages to get, number of all memory pages by default.
    :param bool ignore_missing: if ``True``, ignore missing pages, ``False`` by default.
    :raises ducky.errors.AccessViolationError: when ``ignore_missing == False`` and there's
      a missing page in requested range, this exception is rised.
    :returns: list of pages in area
    :rtype: `list` of :py:class:`ducky.mm.MemoryPage`
    """

    self.DEBUG('mc.pages: pages_start=%s, pages_cnt=%s, ignore_missing=%s', pages_start, pages_cnt, ignore_missing)

    pages_cnt = pages_cnt or self.__pages_cnt

    if ignore_missing is True:
      return (self.__pages[i] for i in range(pages_start, pages_start + pages_cnt) if i in self.__pages)
    else:
      return (self.__pages[i] for i in range(pages_start, pages_start + pages_cnt))

  def pages_in_area(self, address = 0, size = None, ignore_missing = False):
    """
    Return list of memory pages.

    :param u24 address: beggining address of the area, by default 0.
    :param u24 size: size of the area, by default the whole memory size.
    :param bool ignore_missing: if ``True``, ignore missing pages, ``False`` by default.
    :raises ducky.errors.AccessViolationError: when ``ignore_missing == False`` and there's
      a missing page in requested range, this exception is rised.
    :returns: list of pages in area
    :rtype: `list` of :py:class:`ducky.mm.MemoryPage`
    """

    self.DEBUG('mc.pages_in_area: address=%s, size=%s', ADDR_FMT(address), size)

    size = size or self.__size
    pages_start, pages_cnt = area_to_pages(address, size)

    return self.pages(pages_start = pages_start, pages_cnt = pages_cnt, ignore_missing = ignore_missing)

  def boot(self):
    """
    Prepare memory controller for immediate usage by other components.
    """

    # Reserve the first segment for system usage
    self.alloc_segment()

    self.INFO('mm: %s, %s available', sizeof_fmt(self.__size, max_unit = 'Ki'), sizeof_fmt(self.__size - len(self.__pages) * PAGE_SIZE, max_unit = 'Ki'))

  def halt(self):
    for area in list(self.mmap_areas.values()):
      self.unmmap_area(area)

  def update_area_flags(self, address, size, flag, value):
    """
    Update flag of all pages in memory that are inside the specified area.

    :param u24 address: address of the first byte of the area.
    :param int size: number of bytes in the area.
    :param string flag: flag to change. For list of available flags, see
      :py:class:`ducky.mm.MemoryPage`.
    :param bool value: new value of flag.
    """

    self.DEBUG('mc.update_area_flags: address=%s, size=%s, flag=%s, value=%i', address, size, flag, value)

    for pg in self.pages_in_area(address = address, size = size):
      setattr(pg, flag, value)

  def update_pages_flags(self, pages_start, pages_cnt, flag, value):
    """
    Update flag of all pages in specified range.

    :param int pages_start: index of the first page.
    :param int pages_cnt: number of pages.
    :param string flag: flag to change. For list of available flags, see
      :py:class:`ducky.mm.MemoryPage`.
    :param bool value: new value of flag.
    """

    self.DEBUG('mc.update_pages_flags: page=%s, cnt=%s, flag=%s, value=%i', pages_start, pages_cnt, flag, value)

    for pg in self.pages(pages_start = pages_start, pages_cnt = pages_cnt):
      setattr(pg, flag, value)

  def reset_area_flags(self, address, size):
    """
    Reset page flags to default for all pages in memory that are inside
    the specified area.

    :param u24 address: address of the first byte of the area.
    :param int size: number of bytes in the area.
    """

    self.DEBUG('mc.reset_area_flags: address=%s, size=%s', address, size)

    for pg in self.pages_in_area(address = address, size = size):
      pg.flags_reset()

  def reset_pages_flags(self, pages_start, pages_cnt):
    """
    Reset page flags to default for all pages in specified range.

    :param int pages_start: index of the first page.
    :param int pages_cnt: number of pages.
    """

    self.DEBUG('mc.reset_pages_flags: page=%s, size=%s', pages_start, pages_cnt)

    for pg in self.pages(pages_start = pages_start, pages_cnt = pages_cnt):
      pg.flags_reset()

  def __load_content_u8(self, segment, base, content):
    from ..cpu.assemble import SpaceSlot

    bsp  = segment_addr_to_addr(segment, base)
    sp   = bsp
    size = len(content)

    self.DEBUG('mc.__load_content_u8: segment=%s, base=%s, size=%s, sp=%s', UINT8_FMT(segment), UINT16_FMT(base), size, UINT16_FMT(sp))

    for i in content:
      if type(i) == SpaceSlot:
        sp += i.size.u16
      else:
        self.write_u8(sp, i.u8, privileged = True)
        sp += 1

  def __load_content_u16(self, segment, base, content):
    bsp  = segment_addr_to_addr(segment, base)
    sp   = bsp
    size = len(content) * 2

    self.DEBUG('mc.__load_content_u16: segment=%s, base=%s, size=%s, sp=%s', UINT8_FMT(segment), UINT16_FMT(base), size, UINT16_FMT(sp))

    for i in content:
      self.write_u16(sp, i.u16, privileged = True)
      sp += 2

  def __load_content_u32(self, segment, base, content):
    bsp = segment_addr_to_addr(segment, base)
    sp   = bsp
    size = len(content) * 4

    self.DEBUG('mc.__load_content_u32: segment=%s, base=%s, size=%s, sp=%s', UINT8_FMT(segment), UINT16_FMT(base), size, UINT16_FMT(sp))

    for i in content:
      self.write_u32(sp, i.u32, privileged = True)
      sp += 4

  def load_text(self, segment, base, content):
    self.DEBUG('mc.load_text: segment=%s, base=%s', UINT8_FMT(segment), UINT16_FMT(base))

    self.__load_content_u32(segment, base, content)

  def load_data(self, segment, base, content):
    self.DEBUG('mc.load_data: segment=%s, base=%s', UINT8_FMT(segment), UINT16_FMT(base))

    self.__load_content_u8(segment, base, content)

  def __set_section_flags(self, pages_start, pages_cnt, flags):
    self.DEBUG('__set_section_flags: start=%s, cnt=%s, flags=%s', pages_start, pages_cnt, flags)

    self.reset_pages_flags(pages_start, pages_cnt)
    self.update_pages_flags(pages_start, pages_cnt, 'read', flags.readable == 1)
    self.update_pages_flags(pages_start, pages_cnt, 'write', flags.writable == 1)
    self.update_pages_flags(pages_start, pages_cnt, 'execute', flags.executable == 1)

  def load_file(self, file_in, csr = None, dsr = None, stack = True):
    self.DEBUG('mc.load_file: file_in=%s, csr=%s, dsr=%s', file_in, csr, dsr)

    from . import binary

    # One segment for code and data
    csr = csr or self.alloc_segment()
    dsr = dsr or csr
    sp  = None
    ip  = None

    symbols = {}
    regions = []

    with binary.File.open(self.machine.LOGGER, file_in, 'r') as f_in:
      f_in.load()

      f_header = f_in.get_header()

      for i in range(0, f_header.sections):
        s_header, s_content = f_in.get_section(i)

        self.DEBUG('loading section %s', f_in.string_table.get_string(s_header.name))

        s_base_addr = None

        if s_header.type == binary.SectionTypes.SYMBOLS:
          for symbol in s_content:
            symbols[f_in.string_table.get_string(symbol.name)] = UInt16(symbol.address)

          continue

        if s_header.flags.loadable != 1:
          continue

        s_base_addr = segment_addr_to_addr(csr if s_header.type == binary.SectionTypes.TEXT else dsr, s_header.base)
        pages_start, pages_cnt = area_to_pages(s_base_addr, s_header.file_size)

        if f_header.flags.mmapable == 1:
          # Always mmap sections as RW, and disable W if section' flags requires that
          # Otherwise, when program asks Vm to enable W, any access would fail because
          # the underlying mmap area was not mmaped as writable
          self.mmap_area(f_in.name, s_base_addr, s_header.file_size, offset = s_header.offset, flags = s_header.flags, shared = False)

        else:
          for i in range(pages_start, pages_start + pages_cnt):
            self.__alloc_page(i)

          if s_header.type == binary.SectionTypes.TEXT:
            self.load_text(csr, s_header.base, s_content)

          elif s_header.type == binary.SectionTypes.DATA:
            if s_header.flags.bss != 1:
              self.load_data(dsr, s_header.base, s_content)

          self.__set_section_flags(pages_start, pages_cnt, s_header.flags)

        regions.append(MemoryRegion(self, f_in.string_table.get_string(s_header.name), s_base_addr, s_header.file_size, s_header.flags))

    if stack:
      pg, sp = self.alloc_stack(segment = dsr)
      regions.append(MemoryRegion(self, 'stack', pg.base_address, PAGE_SIZE, binary.SectionFlags.create(readable = True, writable = True)))

    return (csr, dsr, sp, ip, symbols, regions, f_in)

  def __get_mmap_fileno(self, file_path):
    if file_path not in self.opened_mmap_files:
      self.opened_mmap_files[file_path] = [0, open(file_path, 'r+b')]

    desc = self.opened_mmap_files[file_path]

    desc[0] += 1
    return desc[1].fileno()

  def __put_mmap_fileno(self, file_path):
    desc = self.opened_mmap_files[file_path]

    desc[0] -= 1
    if desc[0] > 0:
      return

    desc[1].close()
    del self.opened_mmap_files[file_path]

  def mmap_area(self, file_path, address, size, offset = 0, flags = None, shared = False):
    """
    Assign set of memory pages to mirror external file, mapped into memory.

    :param string file_path: path of external file, whose content new area
      should reflect.
    :param u24 address: address where new area should start.
    :param u24 size: length of area, in bytes.
    :param int offset: starting point of the area in mmaped file.
    :param ducky.mm.binary.SectionFlags flags: specifies required flags for mmaped
      pages.
    :param bool shared: if ``True``, content of external file is mmaped as
      shared, i.e. all changes are visible to all processes, not only to the
      current ducky virtual machine.
    :returns: newly created mmap area.
    :rtype: ducky.mm.MMapArea
    :raises ducky.errors.InvalidResourceError: when ``size`` is not multiply of
      :py:data:`ducky.mm.PAGE_SIZE`, or when ``address`` is not multiply of
      :py:data:`ducky.mm.PAGE_SIZE`, or when any of pages in the affected area
      is already allocated.
    """

    self.DEBUG('mc.mmap_area: file=%s, offset=%s, size=%s, address=%s, flags=%s, shared=%s', file_path, offset, size, ADDR_FMT(address), flags.to_string(), shared)

    if size % PAGE_SIZE != 0:
      raise InvalidResourceError('Memory size must be multiple of PAGE_SIZE')

    if address % PAGE_SIZE != 0:
      raise InvalidResourceError('MMap area address must be multiple of PAGE_SIZE')

    pages_start, pages_cnt = area_to_pages(address, size)

    for i in range(pages_start, pages_start + pages_cnt):
      if i in self.__pages:
        raise InvalidResourceError('MMap request overlaps with existing pages: page=%s, area=%s' % (self.__pages[i], self.__pages[i].area))

    mmap_flags = mmap.MAP_SHARED if shared else mmap.MAP_PRIVATE

    # Always mmap as writable - VM will force read-only access using
    # page flags. But since it is possible to change page flags
    # in run-time, and request write access to areas originaly
    # loaded as read-only, such write access would fail because
    # the underlying mmap area was mmaped as read-only only, and this
    # limitation is not possible to overcome.
    mmap_prot = mmap.PROT_READ | mmap.PROT_WRITE

    ptr = mmap.mmap(
      self.__get_mmap_fileno(file_path),
      size,
      flags = mmap_flags,
      prot = mmap_prot,
      offset = offset)

    area = MMapArea(ptr, address, size, file_path, ptr, pages_start, pages_cnt)

    for i in range(pages_start, pages_start + pages_cnt):
      self.register_page(MMapMemoryPage(area, self, i, ptr, offset = (i - pages_start) * PAGE_SIZE))

    self.__set_section_flags(pages_start, pages_cnt, flags)

    self.mmap_areas[area.address] = area

    return area

  def unmmap_area(self, mmap_area):
    self.machine.cpu_cache_controller.release_area_references(self.page(mmap_area.pages_start).base_address, mmap_area.pages_cnt * PAGE_SIZE)
    self.reset_pages_flags(mmap_area.pages_start, mmap_area.pages_cnt)

    for pg in self.pages(pages_start = mmap_area.pages_start, pages_cnt = mmap_area.pages_cnt):
      self.unregister_page(pg)

    del self.mmap_areas[mmap_area.address]

    mmap_area.ptr.close()

    self.__put_mmap_fileno(mmap_area.file_path)

  def read_u8(self, addr, privileged = False):
    self.DEBUG('mc.read_u8: addr=%s, priv=%i', addr, privileged)

    return self.page((addr & PAGE_MASK) >> PAGE_SHIFT).read_u8(addr & (PAGE_SIZE - 1), privileged = privileged)

  def read_u16(self, addr, privileged = False):
    self.DEBUG('mc.read_u16: addr=%s, priv=%i', addr, privileged)

    if self.force_aligned_access and addr % 2:
      raise AccessViolationError('Unable to access unaligned address: addr={}'.format(ADDR_FMT(addr)))

    return self.page((addr & PAGE_MASK) >> PAGE_SHIFT).read_u16(addr & (PAGE_SIZE - 1), privileged = privileged)

  def read_u32(self, addr, privileged = False, not_execute = True):
    self.DEBUG('mc.read_u32: addr=%s, priv=%s, not_execute=%s', addr, privileged, not_execute)

    if self.force_aligned_access and addr % 4:
      raise AccessViolationError('Unable to access unaligned address: addr={}'.format(ADDR_FMT(addr)))

    return self.page((addr & PAGE_MASK) >> PAGE_SHIFT).read_u32(addr & (PAGE_SIZE - 1), privileged = privileged, not_execute = not_execute)

  def write_u8(self, addr, value, privileged = False, dirty = True):
    self.DEBUG('mc.write_u8: addr=%s, value=%s, priv=%i, dirty=%i', addr, value, privileged, dirty)
    self.page((addr & PAGE_MASK) >> PAGE_SHIFT).write_u8(addr & (PAGE_SIZE - 1), value, privileged = privileged, dirty = dirty)

  def write_u16(self, addr, value, privileged = False, dirty = True):
    self.DEBUG('mc.write_u16: addr=%s, value=%s, priv=%i, dirty=%i', addr, value, privileged, dirty)

    if self.force_aligned_access and addr % 2:
      raise AccessViolationError('Unable to access unaligned address: addr={}'.format(ADDR_FMT(addr)))

    self.page((addr & PAGE_MASK) >> PAGE_SHIFT).write_u16(addr & (PAGE_SIZE - 1), value, privileged = privileged, dirty = dirty)

  def write_u32(self, addr, value, privileged = False, dirty = True):
    self.DEBUG('mc.write_u32: addr=%s, value=%s, priv=%i, dirty=%i', addr, value, privileged, dirty)

    if self.force_aligned_access and addr % 4:
      raise AccessViolationError('Unable to access unaligned address: addr={}'.format(ADDR_FMT(addr)))

    self.page((addr & PAGE_MASK) >> PAGE_SHIFT).write_u32(addr & (PAGE_SIZE - 1), value, privileged = privileged, dirty = dirty)

  def save_interrupt_vector(self, table, index, desc):
    """
    save_interrupt_vector(mm.UInt24, int, cpu.InterruptVector)
    """

    from ..cpu import InterruptVector

    self.DEBUG('mc.save_interrupt_vector: table=%s, index=%i, desc=(CS=%s, DS=%s, IP=%s)',
               table, index, desc.cs, desc.ds, desc.ip)

    vector_address = table + index * InterruptVector.SIZE

    self.write_u8(vector_address, desc.cs & 0xFF, privileged = True)
    self.write_u8(vector_address + 1, desc.ds & 0xFF, privileged = True)
    self.write_u16(vector_address + 2, desc.ip & 0xFFFF, privileged = True)

from ..interfaces import IVirtualInterrupt
from ..devices import VIRTUAL_INTERRUPTS, IRQList

class MMInterrupt(IVirtualInterrupt):
  def op_mmap(self, core):
    from .binary import SectionFlags
    from ..cpu.registers import Registers

    core.REG(Registers.R00).value = 0xFFFF

    address = segment_addr_to_addr(core.REG(Registers.DS).value, core.REG(Registers.R01).value)
    size = core.REG(Registers.R02).value
    filepath_ptr = segment_addr_to_addr(core.REG(Registers.DS).value, core.REG(Registers.R03).value)
    offset = core.REG(Registers.R04).value
    flags = core.REG(Registers.R05).value

    filepath = []
    filepath_ptr = segment_addr_to_addr(core.REG(Registers.DS).value, core.REG(Registers.R03).value)

    while True:
      c = core.MEM_IN8(filepath_ptr)
      if c == 0:
        break

      filepath.append(chr(c))

    filepath = ''.join(filepath)

    flags = SectionFlags(readable = flags & MM_FLAG_READ, writable = flags & MM_FLAG_WRITE, executable = flags & MM_FLAG_EXECUTE)

    self.mmap_area(filepath, address, size, offset = offset, flags = flags)

    core.REG(Registers.R00).value = address

  def op_unmmap(self, core):
    from ..cpu.registers import Registers

    core.REG(Registers.R00).value = 0xFFFF

    address = segment_addr_to_addr(core.REG(Registers.DS).value, core.REG(Registers.R01).value)

    area = self.core.memory.mmap_areas.get(address)
    if area is None:
      raise AccessViolationError('Unmmap is not allowed for this area: address=%s' % ADDR_FMT(address))

    self.core.memory.unmmap_area(area)

    core.REG(Registers.R00).value = 0

  def run(self, core):
    from ..cpu import do_log_cpu_core_state
    from ..cpu.registers import Registers

    core.DEBUG('MMInterrupt: triggered')

    do_log_cpu_core_state(core, logger = core.DEBUG)

    op = core.REG(Registers.R00).value

    if op == MMOperationList.MTELL:
      core.REG(Registers.R00).value = 0xFFFF

      start = core.REG(Registers.R01).value
      end   = core.REG(Registers.R02).value
      flags = core.REG(Registers.R03).value

      if start % PAGE_SIZE != 0:
        core.WARN('Start address must be page-aligned: start=%s', ADDR_FMT(start))
        return

      if end % PAGE_SIZE != 0:
        core.WARN('End address must be page-aligned: end=%s', ADDR_FMT(end))
        return

      if end <= start:
        core.WARN('Size of area must be bigger than zero: size=%s', ADDR_FMT(end - start))
        return

      segment = core.REG(Registers.CS).value if flags & MM_FLAG_CS == MM_FLAG_CS else core.REG(Registers.DS).value
      start = segment_addr_to_addr(segment & 0xFF, start)
      end   = segment_addr_to_addr(segment & 0xFF, end)
      size  = end - start

      mc = core.cpu.machine.memory

      flags = 0

      for pg in mc.pages_in_area(address = start, size = size):
        if pg.read:
          flags |= MM_FLAG_READ
        if pg.write:
          flags |= MM_FLAG_WRITE
        if pg.execute:
          flags |= MM_FLAG_EXECUTE
        if pg.dirty:
          flags |= MM_FLAG_DIRTY

      core.REG(Registers.R00).value = 0
      core.REG(Registers.R01).value = 0xFFFF & flags

      core.DEBUG('mtell: start=%s, end=%s, size=%s, flags=%s', ADDR_FMT(start), ADDR_FMT(end), UINT24_FMT(end - start),
                 ''.join([
                   'R' if flags & MM_FLAG_READ else '-',
                   'W' if flags & MM_FLAG_WRITE else '-',
                   'X' if flags & MM_FLAG_EXECUTE else '-',
                   'D' if flags & MM_FLAG_DIRTY else '-',
                   'CS' if flags & MM_FLAG_CS else 'DS'
                 ]))

    elif op == MMOperationList.MPROTECT:
      core.REG(Registers.R00).value = 0xFFFF

      start = core.REG(Registers.R01).value
      end   = core.REG(Registers.R02).value
      flags = core.REG(Registers.R03).value

      core.DEBUG('mprotect: start=%s, end=%s, size=%s, flags=%s', ADDR_FMT(start), ADDR_FMT(end), UINT24_FMT(end - start),
                 ''.join([
                   'R' if flags & MM_FLAG_READ else '-',
                   'W' if flags & MM_FLAG_WRITE else '-',
                   'X' if flags & MM_FLAG_EXECUTE else '-',
                   'D' if flags & MM_FLAG_DIRTY else '-',
                   'CS' if flags & MM_FLAG_CS else 'DS'
                 ]))

      if start % PAGE_SIZE != 0:
        core.WARN('Start address must be page-aligned: start=%s', ADDR_FMT(start))
        return

      if end % PAGE_SIZE != 0:
        core.WARN('End address must be page-aligned: end=%s', ADDR_FMT(end))
        return

      if end <= start:
        core.WARN('Size of area must be bigger than zero: size=%s', ADDR_FMT(end - start))
        return

      segment = core.REG(Registers.CS).value if flags & MM_FLAG_CS == MM_FLAG_CS else core.REG(Registers.DS).value
      start = segment_addr_to_addr(segment & 0xFF, start)
      end   = segment_addr_to_addr(segment & 0xFF, end)
      size  = end - start

      core.cache_controller.release_area_references(start, size)

      core.cpu.machine.memory.reset_area_flags(start, size)

      if flags & MM_FLAG_READ:
        core.cpu.machine.memory.update_area_flags(start, size, 'read', True)

      if flags & MM_FLAG_WRITE:
        core.cpu.machine.memory.update_area_flags(start, size, 'write', True)

      if flags & MM_FLAG_EXECUTE:
        core.cpu.machine.memory.update_area_flags(start, size, 'execute', True)

      if flags & MM_FLAG_DIRTY:
        core.cpu.machine.memory.update_area_flags(start, size, 'dirty', True)

      core.REG(Registers.R00).value = 0

    elif op == MMOperationList.ALLOC:
      core.REG(Registers.R00).value = 0xFFFF

      pages_cnt = core.REG(Registers.R01).value
      segment = core.REG(Registers.DS).value

      core.DEBUG('alloc: pages_cnt=%s, segment=%s', pages_cnt, segment)

      pages = core.cpu.machine.memory.alloc_pages(segment = segment, count = pages_cnt)

      for pg in pages:
        pg.flags_reset()

      core.REG(Registers.R00).value = pages[0].base_address & 0x00FFFF

      core.DEBUG('alloc: address=%s', ADDR_FMT(pages[0].base_address))
      for pg in pages:
        core.DEBUG('alloc: pg=%i', pg.index)

    elif op == MMOperationList.FREE:
      core.REG(Registers.R00).value = 0xFFFF

      address = segment_addr_to_addr(core.REG(Registers.DS).value, core.REG(Registers.R01).value)
      pages_start = addr_to_page(address)
      pages_cnt = core.REG(Registers.R02).value
      size = pages_cnt * PAGE_SIZE

      core.DEBUG('free: address=%s, pages_start=%s, pages_cnt=%s', ADDR_FMT(address), pages_start, pages_cnt)

      core.cache_controller.release_area_references(address, size)
      core.cpu.machine.memory.reset_area_flags(address, size)

      pg = core.cpu.machine.memory.page(pages_start)
      core.cpu.machine.memory.free_pages(pg, pages_cnt)

      core.REG(Registers.R00).value = 0

    elif op == MMOperationList.UNUSED:
      pages = [pg for pg in core.cpu.machine.memory.pages(pages_start = addr_to_page(segment_addr_to_addr(core.REG(Registers.DS).value, 0)), pages_cnt = SEGMENT_SIZE, ignore_missing = True)]

      for pg in pages:
        core.DEBUG('used page: pg=%s', pg)

      core.DEBUG('unused: allocated_pages=%i', len(pages))

      core.REG(Registers.R00).value = SEGMENT_SIZE - len(pages)

    elif op == MMOperationList.MMAP:
      self.op_mmap(core)

    elif op == MMOperationList.UNMMAP:
      self.op_unmmap(core)

    else:
      core.WARN('Unknown mm operation requested: %s', op)
      core.REG(Registers.R00).value = 0xFFFF

VIRTUAL_INTERRUPTS[IRQList.MM.value] = MMInterrupt



from ..import patch

import optparse
import signal
import sys

from six.moves import input

def main():
  from . import add_common_options, parse_options

  parser = optparse.OptionParser()
  add_common_options(parser)

  # Machine configuration
  opt_group = optparse.OptionGroup(parser, 'Machine hardware')
  parser.add_option_group(opt_group)
  opt_group.add_option('--machine-config',
                       dest = 'machine_config',
                       action = 'store',
                       default = None,
                       help = 'Path to machine configuration file')
  opt_group.add_option('--machine-profile',
                       dest = 'machine_profile',
                       action = 'store_true',
                       default = False,
                       help = 'Enable profiling of running binaries')
  opt_group.add_option('--set-option',
                       dest = 'set_options',
                       action = 'append',
                       default = [],
                       metavar = 'SECTION:OPTION=VALUE',
                       help = 'Set option')
  opt_group.add_option('--add-option',
                       dest = 'add_options',
                       action = 'append',
                       default = [],
                       metavar = 'SECTION:OPTION=VALUE')
  opt_group.add_option('--set-device-option',
                       dest = 'set_device_options',
                       action = 'append',
                       default = [],
                       metavar = 'DEVICE:OPTION=VALUE',
                       help = 'Set device options')
  opt_group.add_option('--add-device-option',
                       dest = 'add_device_options',
                       action = 'append',
                       default = [],
                       metavar = 'DEVICE:OPTION=VALUE')
  opt_group.add_option('--enable-device',
                       dest = 'enable_devices',
                       action = 'append',
                       default = [],
                       metavar = 'DEVICE',
                       help = 'Enable device')
  opt_group.add_option('--disable-device',
                       dest = 'disable_devices',
                       action = 'append',
                       default = [],
                       metavar = 'DEVICE',
                       help = 'Disable device')

  # Debug options
  opt_group = optparse.OptionGroup(parser, 'Debug options')
  parser.add_option_group(opt_group)
  opt_group.add_option('-g', '--go-on',
                       dest = 'go_on',
                       action = 'store_true',
                       default = False,
                       help = 'Don\'t wait for user to start binaries with pressing Enter')
  opt_group.add_option('-p', '--profile',
                       dest = 'profile',
                       action = 'store_true',
                       default = False,
                       help = 'Enable profiling of the whole virtual machine')
  opt_group.add_option('-P', '--profile-dir',
                       dest = 'profile_dir',
                       action = 'store',
                       default = None,
                       help = 'Store profiling data in this directory')
  opt_group.add_option('--debug-open-files', dest = 'debug_open_files', action = 'store_true', default = False, help = 'List all open - not closed cleanly - files when VM quits')
  opt_group.add_option('--stdio-console',    dest = 'stdio_console',    action = 'store_true', default = False, help = 'Enable console terminal using stdin/stdout as IO streams')

  options, logger = parse_options(parser)

  if options.machine_config is None:
    parser.print_help()
    sys.exit(1)

  from ..profiler import STORE

  if options.profile:
    STORE.enable_machine()

  if options.machine_profile:
    STORE.enable_cpu()

  main_profiler = STORE.get_machine_profiler()
  main_profiler.enable()

  from ..machine import Machine
  M = Machine()

  if options.stdio_console:
    from ..console import TerminalConsoleConnection
    console_slave = TerminalConsoleConnection(0, M.console)
    console_slave.boot()
    M.console.connect(console_slave)

  from ..config import MachineConfig
  machine_config = MachineConfig()

  if options.machine_config is not None:
    machine_config.read(options.machine_config)

  for option in options.set_options:
    section, option = option.split(':')

    if not machine_config.has_section(section):
      logger.error('Unknown config section %s', section)
      continue

    option, value = option.split('=')

    machine_config.set(section, option, value)

  for option in options.add_options:
    section, option = option.split(':')

    if not machine_config.has_section(section):
      logger.error('Unknown config section %s', section)
      continue

    option, value = option.split('=')
    machine_config.set(section, option, machine_config.get(section, option) + ', ' + value)

  for option in options.set_device_options:
    dev, option = option.split(':')

    if not machine_config.has_section(dev):
      logger.error('Unknown device %s', dev)
      continue

    option, value = option.split('=')

    machine_config.set(dev, option, value)

  for option in options.add_device_options:
    dev, option = option.split(':')

    if not machine_config.has_section(dev):
      logger.error('Unknown device %s', dev)
      continue

    option, value = option.split('=')
    machine_config.set(dev, option, machine_config.get(dev, option) + ', ' + value)

  for dev in options.enable_devices:
    machine_config.set(dev, 'enabled', True)

  for dev in options.disable_devices:
    machine_config.set(dev, 'enabled', False)

  M.hw_setup(machine_config)

  def signal_handler(sig, frame):
    if sig == signal.SIGUSR1:
      logger.info('VM suspended by user')
      M.reactor.add_call(M.suspend)

    elif sig == signal.SIGUSR2:
      logger.info('VM unsuspended by user')
      M.reactor.add_call(M.wake_up)

    elif sig == signal.SIGINT:
      logger.info('VM halted by user')
      M.reactor.add_call(M.halt)

    elif sig == signal.SIGSEGV:
      logger.info('VM snapshot requested')
      M.reactor.add_call(M.snapshot('ducky-snapshot-user.bin'))

  signal.signal(signal.SIGINT,  signal_handler)
  signal.signal(signal.SIGUSR1, signal_handler)
  signal.signal(signal.SIGUSR2, signal_handler)
  signal.signal(signal.SIGSEGV, signal_handler)

  M.boot()

  if not options.go_on:
    input('Press Enter to start execution of loaded binaries')

  M.run()  # reactor loop!

  table_exits = [
    ['Core', 'Exit code']
  ]

  table_inst_caches = [
    ['Core', 'Reads', 'Inserts', 'Hits', 'Misses', 'Prunes']
  ]
  table_data_caches = [
    ['Core', 'Reads', 'Hits', 'Misses', 'Prunes', 'Forced writes']
  ]
  table_cnts = [
    ['Core', 'Ticks']
  ]

  def __check_stats(core):
    table_exits.append([str(core), core.exit_code])

    table_inst_caches.append([
      str(core),
      core.instruction_cache.reads,
      core.instruction_cache.inserts,
      core.instruction_cache.hits,
      core.instruction_cache.misses,
      core.instruction_cache.prunes
    ])

    if core.data_cache is not None:
      table_data_caches.append([
        str(core),
        core.data_cache.reads,
        core.data_cache.hits,
        core.data_cache.misses,
        core.data_cache.prunes,
        core.data_cache.forced_writes
      ])

    table_cnts.append([
      str(core),
      core.registers.cnt.value
    ])

  for core in M.cores():
    __check_stats(core)

  logger.info('')
  logger.info('Exit codes')
  logger.table(table_exits)
  logger.info('')
  logger.info('Instruction caches')
  logger.table(table_inst_caches)
  logger.info('')
  logger.info('Data caches')
  logger.table(table_data_caches)
  logger.info('')
  logger.table(table_cnts)
  logger.info('')

  inst_executed = sum([core.registers.cnt.value for core in M.cores()])
  runtime = float(M.end_time - M.start_time)
  logger.info('Executed instructions: %i %f (%.4f/sec)', inst_executed, runtime, float(inst_executed) / runtime)
  logger.info('')

  main_profiler.disable()

  if options.profile or options.machine_profile:
    logger.info('Saving profiling data into %s' % options.profile_dir)
    STORE.save(options.profile_dir)

  sys.exit(M.exit_code)



import enum

from ..interfaces import IMachineWorker
from ..mm import UINT16_FMT

VIRTUAL_INTERRUPTS = {}

class IRQList(enum.IntEnum):
  """
  List of known IRQ sources.
  """

  # HW devices
  TIMER    = 0
  KEYBOARD = 1

  # SW interrupts and exceptions
  HALT    = 32
  BLOCKIO = 33
  VMDEBUG = 34
  MM      = 35
  MATH    = 36

  IRQ_COUNT = 64


class IOPorts(enum.IntEnum):
  PORT_COUNT = 65536


class Device(IMachineWorker):
  def __init__(self, machine, klass, name, *args, **kwargs):
    super(Device, self).__init__()

    self.machine = machine
    self.klass = klass
    self.name = name

    self.master = None

  @staticmethod
  def create_from_config(machine, config, section):
    return None

  def is_slave(self):
    return self.master is not None

  def get_master(self):
    return self.master


class IRQProvider(object):
  pass


class IOProvider(object):
  def is_port_protected(self, port, read = True):
    return False

  def read_u8(self, port):
    self.machine.WARN('Unhandled port: %s', UINT16_FMT(port))

    return None

  def read_u16(self, port):
    self.machine.WARN('Unhandled port: %s', UINT16_FMT(port))

    return None

  def write_u8(self, port, value):
    self.machine.WARN('Unhandled port: %s', UINT16_FMT(port))

  def write_u16(self, port, value):
    self.machine.WARN('Unhandled port: %s', UINT16_FMT(port))
